"""
   Cars of various kinds

   Pyweek 5 Scavengers team entry  Sep 2007

   GPL v3 or later
"""

from math import radians, cos, sin, degrees, atan2, fmod
from OpenGL.GL import *
from tdlib.vec import Vec
from tdlib import hox, part
import data
import objpart
import materials
from racetrack import NegativeHuge

car_materials = [
    "Red","Orange","Yellow","Green","Blue","Purple","White",
    "Steel","Gold","Chocolate","Shadow","Water","Slime"]

class Car(part.Part):
    _default_style={'obj-filename':'box.obj','override-mtl':'White',
                    'mtl-override-pieces':['box'],
                    'key-filename':'key.obj'}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':1.0,'keyRY':0.0,
                   'keyRZ':0.0,'keyTX':0.0}
    _style_attributes=('obj-filename','opaque-pieces',
                       'mtl-override-pieces','override-mtl')
    top_speed = 0.4
    max_accel = 0.03
    brake_friction = 0.5
    ground_friction = 0.9
    handling = 0.5
    turn_angle = 2.0
    wind_effect = 0.1
    thrust = 0.0
    velocity = Vec(0,0,0)
    flung = False
    braking = False
    gone = False
    steering = 0
    radius = 0.7
    keyturn = 1.0

    """ A vehicle (or similar) """
    def __init__(self,name='',**kwd):
        super(Car,self).__init__(name,**kwd)
        self.park()
        self.lap = []
        self.lap_number = 1
        self.running = True
        self.hox_opaque = []
        self.hox_transparent = []
        self.piece_dict = {}

    def park(self):
        self.thrust = 0.0
        self.velocity = Vec(0,0,0)
        self.angle = 0.0
        self.flung = False
        self.braking = False
        self.gone = False
        self.steering = 0
        
    def turn(self, steering=0):
        """Apply turning left (steering=-1), right (+1) or ahead (0)"""
        if self.flung:
            # Out of control!
            #self.angle += self.turn_angle * self.steering * 2.0
            return
        self.angle -= (self.turn_angle * steering)
        self.steering = steering

    def brake(self,braking):
        self.braking = braking
        if braking:
            self.thrust = 0.0
        else:
            self.thrust = self.max_accel

    def drive(self, track, wind_vec):
        """ Apply a change to car's speed and position

        if flung, ignore thrust, braking etc. Just do flying...
        find steering direction.
        if speed along steering direction >= top_speed, ignore thrust
        thrust along steering direction is otherwise added to velocity
        if speed along wind_vec >= length of wind_vec, ignore wind
        wind is wind_vec times wind_effect, added to velocity

        position is updated by velocity
        if position goes off-track, car is "flung"
        """
        v = self.velocity
        if not self.flung:
            a = radians(self.angle)
            steering = Vec(cos(a), sin(a),0.0)
            speed_along_steering = max(v.dot(steering),0.0)
            # forward thrust
            thrust = min(self.thrust, self.top_speed - speed_along_steering)
            v += steering * thrust
            # steering
            vn = v.orthoXY().normalise()
            vz = v.z
            b = vn.dot(steering) # cos(angle between velocity & steering)
            if b > (1 - self.handling):
                k = b * self.handling
                v = steering * k + v * (1 - k)
            # braking
            if self.braking:
                v *= self.brake_friction
            v.z = vz

        wind_direction = wind_vec.normalise()
        speed_along_wind = v.dot(wind_direction)
        wind_speed = wind_vec.length()
        relative_wind_speed = max(wind_speed - speed_along_wind,0.0)
        wind_thrust = relative_wind_speed * self.wind_effect
        v += wind_direction * wind_thrust

        z_here = self.pos[2]
        pos = v + self.pos
        x = pos.x
        y = pos.y
        z = pos.z
        ground = track.z_at(x,y)

        collider = track.first_collision(self,x,y,z,self.radius)
        if collider:
            # maximally stupid bounce algorithm:
            # drop velocity to zero plus an amount away from obstacle
            speed = v.length()
            v = (Vec(x,y,z) - collider.pos).normalise() * speed

        # Gravity
        if z + v.z > ground:
            v.z -= 0.01 # falling
            self.flung = True
        else:
            if (ground - z) > 1.0:
                v *= 0.5    # slow down a lot on up-slopes
            z = ground
            v.z = 0.0
            # friction with the ground
            v *= self.ground_friction
            if z == NegativeHuge:
                self.gone = True
            else:
                self.flung = False
                if ground - z_here > 1.0:
                    pos = self.pos # can't move too far uphill
        
        self.pos = (x,y,z)
        self.velocity = v

    def auto_steer(self,track):
        if not self.lap:
            return
        coords = self.lap[0]
        if len(coords) != 2:
            print coords
            return
        tx,ty = coords
        mx,my = track.tile_xy(tx,ty)
        tw = track.tilewidth * 0.5
        mx,my = mx+tx, my+tw
        x,y,z = self.pos
        theta = degrees(atan2(my-y, mx-x)) - fmod(self.angle,360.0)
        if theta > 180:
            theta = theta - 360
        elif theta < -180:
            theta = theta + 360
    
        if theta <= -self.turn_angle:
            steering = 1
        elif theta >= self.turn_angle:
            steering = -1
        else:
            steering = 0
        need_brake = (abs(theta) > 30 * self.turn_angle)
        self.brake(need_brake)
        self.turn(steering)

    def overlaps(self,(x,y,z),height,radius):
        """Car quacks like a floorset.CylinderRegion"""
        px,py,pz = self.pos
        ztest = (z-1.0) <= pz < (z+height)
        if ztest:
            pos = Vec(self.pos)
            centre = pos.projXY()
            v = centre - (x,y)
            d2 = v.dot(v)
            reach = self.radius + radius
            if d2 > (reach * reach):
                return False
        return ztest

    def prepare(self):
        """ Prepare hoxlist and WFObj from style """
        self.hox_opaque = []
        self.hox_trans = []
        fname = self.getstyle("obj-filename")
        if fname:
            self.obj = objpart.get_obj(fname)
        else:
            self.obj = None
            print self.__class__.__name__,self._name,"failed to load",fname 
        fname = self.getstyle("key-filename")
        if fname:
            self.key_obj = objpart.get_obj(fname)
        else:
            self.key_obj = None
            print self.__class__.__name__,self._name,"failed to load",fname 
        self.piece_dict = {}
        obj_pieces = self.getstyle('obj-pieces')
        if obj_pieces is None:
            obj_pieces = self.obj.pieces()
        # Build lists for hox.process
        opaque_pieces = set(self.getstyle("opaque-pieces",()))
        override_pieces = set(self.getstyle("mtl-override-pieces",()))
        override_mtl = self.getstyle('override-mtl',None)
        override_dl = materials.get(override_mtl)

        self._has_transparent = materials.is_transparent(override_mtl)
        
        for piece in obj_pieces:
            dl = (override_dl if piece in override_pieces else None)
            self.piece_dict[piece] = (self.obj,piece,dl)
            self.hox_opaque.append((hox.DRAW,piece))
            if piece not in opaque_pieces or self.obj.is_transparent(piece):
                self.hox_trans.append((hox.DRAW,piece))
                self._has_transparent = True

        if self.key_obj:
            klist = [(hox.ROT,"keyRY",hox.Y,1.0),
                     (hox.MOVE,"keyTZ",hox.Z,1.0),
                     (hox.MOVE,"keyTX",hox.X,1.0),
                     (hox.ROT,"keyRZ",hox.Z,1.0)]
            for piece in self.key_obj.pieces():
                self.piece_dict["Key:" + piece] = (self.key_obj,piece,0)
                klist.append((hox.DRAW,"Key:"+piece))
            self.hox_opaque.append(klist)
            self.hox_trans.append(klist)

    def drawpiece(self,piece):
        """ Draw a piece of either the car or the key, or any display list"""
        obj,pname,dl = self.piece_dict.get(piece,(None,None,None))
        if dl:
            glCallList(dl)
        if obj:
            obj.drawpiece(pname)

    def render(self,mode):
        glEnable(GL_CULL_FACE)
        if mode == 'OPAQUE':
            glCullFace(GL_BACK)
            hox.process(self.hox_opaque,self.getgeom,self.drawpiece)
        elif mode == 'TRANSPARENT':
            for cull in (GL_FRONT, GL_BACK):
                glCullFace(cull)
                hox.process(self.hox_trans,self.getgeom,self.drawpiece)
        elif mode == 'PICK':
            glCullFace(GL_BACK)
            picking.label((self.__class__.__name__, self._name))
            hox.process(self.hox_opaque,self.getgeom,self.drawpiece,False)
            picking.nolabel()

    def setup_style(self):
        glPushAttrib(GL_LIGHTING_BIT)
        glEnable(GL_LIGHTING)

    def setdown_style(self):
        glPopAttrib()

    def step(self):
        if self.running:
            self._geom["keyRZ"] += self.keyturn

class Boat(Car):
    _default_style={'obj-filename':'boat.obj','override-mtl':'Blue',
                    'mtl-override-pieces':['Cylinder'],'key-filename':'key.obj'}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':1.5,'keyRY':-90.0,
                   'keyRZ':0.0,'keyTX':0.5}
    top_speed = 1.0
    max_accel = 0.15
    brake_friction = 0.2
    turn_angle = 2.5
    handling = 0.8
    wind_effect = 0.1

class Racer(Car):
    _default_style={'obj-filename':'car.obj','override-mtl':'Red',
                    'mtl-override-pieces':['Body_Cube'],'key-filename':'key.obj'}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':1.1,'keyRY':0.0,
                   'keyRZ':0.0,'keyTX':-0.1}
    top_speed = 0.6
    max_accel = 0.05
    brake_friction = 0.4
    turn_angle = 2.2
    handling = 0.7
    wind_effect = 0.05

class OnionCar(Car):
    _default_style={'obj-filename':'onion.obj','override-mtl':'Orange',
                    'mtl-override-pieces':["Wheels_Cylinder"],'key-filename':'key.obj'}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':0.5,'keyRY':-80.0,
                   'keyRZ':0.0,'keyTX':0.4}
    top_speed = 0.5
    max_accel = 0.08
    brake_friction = 0.6
    turn_angle = 4.0
    handling = 0.9
    wind_effect = 0.1

class Crocodile(Car):
    _default_style={'obj-filename':'crocodile.obj',
                    'override-mtl':'Green','key-filename':'key.obj',
                    'mtl-override-pieces':['Cylinder','Cube','Cube.001','Cube.002']}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':0.6,'keyRY':0.0,
                   'keyRZ':0.0,'keyTX':0.0}
    top_speed = 0.6
    max_accel = 0.07
    brake_friction = 0.45
    turn_angle = 2.5
    handling = 0.95
    wind_effect = 0.02

class Helicopter(Car):
    _default_style={'obj-filename':'helicopter.obj',
                    'override-mtl':'Purple','key-filename':'rotor.obj',
                    'mtl-override-pieces':['Body_Sphere']}
    _default_geom={'pos':(0.0,0.0,0.0),'angle':0.0,'keyTZ':1.0,'keyRY':0.0,
                   'keyRZ':0.0,'keyTX':0.0}
    top_speed = 0.8
    max_accel = 0.05
    brake_friction = 0.6
    turn_angle = 2.5
    handling = 0.95
    wind_effect = 0.2
    def brake(self,braking):
        self.flung = False
        super(Helicopter,self).brake(braking)

all_cars = [
    ('Racing Car',Racer),
    ('Boat',Boat),
    ('Crocodile',Crocodile),
    ('Onion',OnionCar),
    ('Box Car',Car),
    ('Helicopter',Helicopter)
    ]
