"""
  gamescreens

      TitleScreen
      OptionsScreen
      CarScreen
      TrackScreen
      RaceScreen
      
   Pyweek 5 Scavengers team entry  Sep 2007

   GPL v3 or later

"""
from __future__ import division
from math import radians, sin, cos, fmod, atan2, degrees

import time
import random

import pygame
from pygame.locals import *
from tdlib import (
    part, mainloop, usualGL, camera, viewgroup, stylesheet,
    backdrop, animator, textpanel, vec, texture, lighting, picking
    )
from OpenGL.GL import *

from functools import partial

import data
import cars
import sky
import tornado
import racetrack
import materials
import objpart
import credits

default_stylesheet = {
    "Backdrop":{"background-colour":(0.5,0.5,0.8,1.0)},
    "TextPanel":{
        'colour':(1,1,1,1),
        'background-colour':(0.0,0.0,0.0,0.0),
        'border-colour':(1,1,1,1), 'border-width':0,
        'highlight-background':(0,0,1,1), 'pitch':10,
        'texture':None, 'texture-repeat':None,
        'padding':0.8},
    '#left':dict(image='K_LEFT.png'),
    '#right':dict(image='K_RIGHT.png'),
    }

title_stylesheet = {
    "Banner":{"image":"title.png"},
    "Tornado":{ "texture":"twister-dark.png" },
    "SkyBall":{ "texture":"stormy.jpg"},
    "ViewGroup":{"light-pos":(4,-4,3,0),
                 "light-colour":(1,1,0.8,1)},
    "#start":{'obj-filename':'start.obj'},
    "#credits":{"pitch":20, "colour":(1,1,0,0.5),"font":"Courier",
                "border-width":0},
    }

cars_stylesheet = {
    "Backdrop":{"background-colour":(0.5,0.5,0.8,1.0)},
    "ViewGroup":{"light-pos":(6,1,1,1),
                 "light-colour":(1,1,1,1)},
    }

tracks_stylesheet = {
    "Backdrop":{"background-colour":(0.2,0.5,0.4,1.0)},
    "ViewGroup":{"light-pos":(6,1,1,1),
                 "light-colour":(1,1,1,1)},
    }

colours_stylesheet = {
    "Backdrop":{"background-colour":(0.3,0.3,0.3,1.0)},
    "ViewGroup":{"light-pos":(6,1,1,1),
                 "light-colour":(1,1,1,1)},
    }

race_stylesheet = {
    "Tornado":{ "texture":"twister-dark.png",'colour':(1,1,1,0.3) },
    "SkyBall":{ "texture":"stormy.jpg"},
    "SkyBall.clear": {"texture":"clear.jpg"},
    "ViewGroup":{"light-pos":(-10,-4,8,0),
                 "light-colour":(0.4,0.4,0.3,1)},
    "#marker":{'obj-filename':'arrow.obj', 'mtl-override-pieces':['Cube'],
               'override-mtl':'TransYellow'},
    "#start":{'obj-filename':'start.obj'},
    "#to_marker":{'colour':(1,1,0,0.5),"texture":"course_pointer.png"},
    "#wind_dir":{'colour':(0.5,0.8,1,0.5),"texture":"wind_pointer.png"},
    "TextPanel":{'pitch':12,'colour':(0.8,1.0,0.8,0.5),
                 'background-colour':(0.0,0.0,0.0,0.0),'border-width':0},
    }

def init(size, fullscreen):
    mainloop.open_window(size,fullscreen)
    usualGL.usual_setup()
    glDisable(GL_COLOR_MATERIAL)
    stylesheet.load(default_stylesheet)
    pygame.display.set_caption("Vortex Racers - PyWeek5 - Scavengers Team")
    pygame.display.set_icon(pygame.image.load(data.filepath("icon.png")))

class Banner(part.Part):
    """A simple banner - one textured quad flat in XY plane"""
    _default_style = {'image':'title.png'}
    _style_attributes = ('image',None)
    _default_geom = dict(left=-1.3, bottom=0.0, width=2.0, height=1.0)
    def __init__(self,name,**kw):
        super(Banner,self).__init__(name,**kw)
        self.tex = None
    def prepare(self):
        imgname = self.getstyle("image")
        self.tex = texture.get_texture(data.filepath(imgname))
        self._has_transparent = self.tex.transparent
    def setup_style(self):
        glPushAttrib(GL_LIGHTING_BIT)
        glDisable(GL_LIGHTING)
        glColor4f(1,1,1,1)
        glEnable(GL_TEXTURE_2D)
        self.tex.bind()
    def setdown_style(self):
        glPopAttrib()
    def render(self,mode):
        if mode == 'PICK':
            picking.label((self.__class__.__name__,self._name))
        x = self.getgeom("left",0)
        y = self.getgeom("bottom",0)
        width = self.getgeom("width",2.0)
        height = self.getgeom("height",1.0)
        glv = glVertex2f
        glt = glTexCoord2i
        glBegin(GL_QUADS)
        glt(0,0); glv(x,y)
        glt(1,0); glv(x+width,y)
        glt(1,1); glv(x+width,y+height)
        glt(0,1); glv(x,y+height)
        glEnd()
        if mode == 'PICK':
            picking.nolabel()

def standard_ortho_geom():
    geom = dict(left=-1.0,right=1.0,top=1.0,bottom=-1.0,near=-1.0,far=100.0)
    swidth,sheight = pygame.display.get_surface().get_size()
    ratio = swidth / sheight
    geom['left'] *= ratio
    geom['right'] *= ratio
    return geom
    
class TitleScreen(part.Group):
    def __init__(self,name="",**kw):
        super(TitleScreen,self).__init__(name,contents=[],**kw)
        self.choice = None
        camera.look_at((-20,0,0))
        camera.look_from((0,-200,-20))
        camera.choose_up((0,0,1))
        skyball=sky.SkyBall('sky')
        twister=tornado.Tornado('twister')
        start = objpart.ObjPart('start')
        start.pos = (-20,-100,-20)
        start.angle = 45.0
        vgroup = viewgroup.ViewGroup('view',[skyball,twister,start])
        self.append(vgroup)
        geom = standard_ortho_geom()
        banner = Banner('banner')
        creds = credits.CreditsRoll("credits")
        creds.pos = (0.5, -0.5, 0)
        ogroup = viewgroup.OrthoGroup('ortho',[banner,creds],geom=geom)
        self.append(ogroup)

    def event(self, event):
        choice = None
        if event.type == QUIT:
            self.choice = "Quit"
            self._expired = True
        elif event.type == KEYDOWN:
            choice = {K_o:"Options",
                      K_p:"Play",
                      K_s:"Play",
                      K_RETURN:"Play",
                      K_ESCAPE:"Quit",
                      K_q:"Quit"}.get(event.key)
        elif event.type == MOUSEBUTTONDOWN and event.button == 1:
            x,y = event.pos
            y = pygame.display.get_surface().get_height() - y
            picking.start(x,y,1,1,True)
            self["view"].draw('PICK')
            objects = picking.end()
            if objects:
                minz,maxz,label = objects[0]
                if label and label[0] == "ObjPart" and label[1] == "start":
                    choice = "Play"
        if choice:
            self.choice = choice
            camera.look_at((0,0,0),10)
            camera.look_from((0,-1,0),60)

    def expired(self):
        return self.choice and camera.ready()

class Carousel(part.Group):
    _default_geom={'radius':5.0, 'angle':0.0, 'pos':(0.0,0.0,0.0)}
    def __init__(self,name,contents,**kw):
        super(Carousel,self).__init__(name,contents,**kw)
        self.mutator = animator.Mutator(self._geom)
        self.choice = 0
        self.angle = 0.0
        self.arrange()
    def arrange(self):
        r = self.getgeom('radius',5.0)
        for i,p in enumerate(self.contents):
            a = (360.0 * i) / len(self.contents)
            p.angle = a + 90.0
            p.pos = (r*cos(radians(a)), r*sin(radians(a)), 0.0)
    def step(self):
        self.mutator.step()
        self.step_contents()

    def rot_left(self):
        n = len(self.contents)
        if n:
            self.choice += 1
            a = -360*self.choice / n
            self.mutator.change('angle', a, 60)
    def rot_right(self):
        n = len(self.contents)
        if n:
            self.choice -= 1
            a = -360*self.choice / n
            self.mutator.change('angle', a, 60)
            
    def get_choice(self):
        return self.choice % len(self.contents) if self.contents else 0

class ChooserScreen(part.Group):
    optname = "choice"
    chosen = False
    quit = False
    def __init__(self,name,options,**kw):
        self.options = options
        super(ChooserScreen,self).__init__(name,contents=(),**kw)
        self.build_parts()
        
    def build_parts(self):
        camera.look_at((5,0,0))
        camera.look_from((20,0,5))
        choices = self.build_choices()
        carousel = self.carousel = Carousel('carousel',choices)
        carousel.pos = (0,0,-1.0)
        back = backdrop.Backdrop('back')
        self.append(back)
        vgroup = viewgroup.ViewGroup('view',[carousel])
        self.append(vgroup)
        geom = standard_ortho_geom()
        panel = self.namepanel = textpanel.TextPanel('namepanel',["*Name*"])
        panel.pos = (0,-0.7,0)
        lefticon = Banner('left',geom=dict(left=-1.2, bottom=-0.8, width=0.2, height=0.2))
        righticon = Banner('right',geom=dict(left=1.0, bottom=-0.8, width=0.2, height=0.2))
        ogroup = viewgroup.OrthoGroup('ortho',[panel,lefticon,righticon],geom=geom)
        self.append(ogroup)
        self.show_choice()

    def build_choices(self):
        return []

    def rot_left(self):
        self.carousel.rot_left()
        self.show_choice()
        
    def rot_right(self):
        self.carousel.rot_right()
        self.show_choice()
        
    def event(self,event):
        if event.type == QUIT:
            self.quit = True
        elif event.type == KEYDOWN:
            if event.key == K_LEFT:
                self.rot_right()
            elif event.key == K_RIGHT:
                self.rot_left()
            elif event.key == K_RETURN:
                self.chosen = True
                self.options[self.optname] = self.get_choice()
            elif event.key == K_ESCAPE:
                self.quit = True

    def get_choice(self):
        carousel = self['carousel']
        return carousel.contents[carousel.get_choice()]

    def show_choice(self):
        self.namepanel[0] = self.get_choice()._name
        
    def expired(self):
        return self.chosen or self.quit

class CarScreen(ChooserScreen):
    optname = "car"
    choice_attr = "_name"
    def __init__(self,options,**kw):
        super(CarScreen,self).__init__("carscreen",options,**kw)

    def build_choices(self):
        car_choices = []
        for name, fn in cars.all_cars:
          car_choices.append(fn(name))
        return car_choices        

class TrackScreen(ChooserScreen):
    optname = "track"
    choice_attr = "title"
    def __init__(self,options,**kw):
        super(TrackScreen,self).__init__("trackscreen",options,**kw)

    def build_choices(self):
        track_choices = []
        for track in racetrack.all_tracks():
            track.topo = None # not needed for thumbnails
            track_choices.append(racetrack.LandscapeThumb(track.title, track))
        return track_choices

    def get_choice(self):
        carousel = self['carousel']
        thumb = carousel.contents[carousel.get_choice()]
        return thumb.track

    def show_choice(self):
        self.namepanel[0] = self.get_choice().title

class ColourScreen(ChooserScreen):
    optname="material"
    def __init__(self,options,**kw):
        super(ColourScreen,self).__init__("colourscreen",options,**kw)
    def build_choices(self):
        self.car = self.options["car"]
        self.mat_list = cars.car_materials
        try:
            self.mat_num = self.mat_list.index(self.car.getstyle("override-mtl"))
        except ValueError:
            self.mat_num = 0
        return [self.car]
    def get_choice(self):
        carousel = self['carousel']
        car = carousel.contents[carousel.get_choice()]
        return car.getstyle("override-mtl")
    def show_choice(self):
        self.namepanel[0] = self.get_choice()
    def rot_left(self):
        self.mat_num = (self.mat_num-1) % len(self.mat_list)
        car = self.car
        car._style["override-mtl"] = self.mat_list[self.mat_num]
        car.prepare()
        self.carousel.rot_left()
        self.show_choice()
    def rot_right(self):
        self.mat_num = (self.mat_num+1) % len(self.mat_list)
        car = self.car
        car._style["override-mtl"] = self.mat_list[self.mat_num]
        car.prepare()
        self.carousel.rot_right()
        self.show_choice()

class RankingsScreen(part.Group):
    def __init__(self,options,**kw):
        super(RankingsScreen,self).__init__("rankscreen",contents=[],**kw)
        self._expired = True #STUB
        
    def event(self,event):
        if event.type == QUIT:
            self._expired = True

class OptionsScreen(part.Group):
    def __init__(self,options,**kw):
        super(OptionsScreen,self).__init__("optscreen",contents=[],**kw)
        self._expired = True #STUB
        
    def event(self,event):
        if event.type == QUIT:
            self._expired = True


class PointerDisplay(part.ScalePart):
    """Show direction of next marker, relative to car's angle, or
       wind direction, or whatever"""
    _default_geom = dict(pos=(1.0,-0.5,0),scale=0.3,angle=0.0)
    _default_style = dict(colour=(1,1,1,0.5),texture="pointer.png")
    _style_attributes = ("colour","texture")
    _has_transparent = True
    def __init__(self,name,**kw):
        super(PointerDisplay,self).__init__(name,**kw)
        self.tex = None
    def prepare(self):
        tfile = self.getstyle("texture")
        if tfile:
            self.tex = texture.get_texture(data.filepath(tfile))
        self.colour = self.getstyle("colour")
    def setup_style(self):
        if self.tex:
            self.tex.bind()
        glColor4fv(self.colour)
        glPushAttrib(GL_LIGHTING_BIT)
        glDisable(GL_LIGHTING)
        glEnable(GL_TEXTURE_2D)
    def setdown_style(self):
        glPopAttrib()
    def render(self,mode):
        glt = glTexCoord2i
        glv = glVertex2f
        glBegin(GL_QUADS)
        for s,t in [(0,0),(1,0),(1,1),(0,1)]:
            glt(s,t)
            glv(s-0.5,t-0.5)
        glEnd()
    def point_along(self,dx,dy,car_angle):
        self.angle = degrees(atan2(dy,dx)) - car_angle
        
        
class RaceScreen(part.Group):
    def __init__(self,options,**kw):
        self.options = options
        super(RaceScreen,self).__init__("race",contents=[],**kw)
        self.marker_light = lighting.claim_light()
        pygame.mixer.music.load(data.filepath("sound/wind.ogg"))
        pygame.mixer.music.set_volume(0.1)
        self.clock = pygame.time.Clock()
        self.one_second = 60.0
        self.build_parts()
        self.won = False
        self.cars_finished = 0
        self.quit = False
        self.laps = 0
        self.steering_left = False
        self.steering_right = False
        self.auto_steering = False
        self.braking = False
        self.racing = False
        self.race_start_time = None
        self.viewpoint = "car"
        self.paused = False
        self.keymap = {
            (KEYDOWN,K_ESCAPE):self.escape_key,
            (KEYDOWN,self.options["key_left"]):partial(self.steer,True,True),
            (KEYDOWN,self.options["key_right"]):partial(self.steer,False,True),
            (KEYDOWN,self.options["key_brake"]):partial(self.brake,True),
            (KEYUP,self.options["key_left"]):partial(self.steer,True,False),
            (KEYUP,self.options["key_right"]):partial(self.steer,False,False),
            (KEYUP,self.options["key_brake"]):partial(self.brake,False),
            (KEYDOWN,K_LEFT):partial(self.steer,True,True),
            (KEYDOWN,K_RIGHT):partial(self.steer,False,True),
            (KEYDOWN,K_DOWN):partial(self.brake,True),
            (KEYUP,K_LEFT):partial(self.steer,True,False),
            (KEYUP,K_RIGHT):partial(self.steer,False,False),
            (KEYUP,K_DOWN):partial(self.brake,False),
            (KEYDOWN,K_F1):partial(self.choose_viewpoint,"car"),
            (KEYDOWN,K_F2):partial(self.choose_viewpoint,"inside"),
            (KEYDOWN,K_F3):partial(self.choose_viewpoint,"bird"),
            (KEYDOWN,K_F4):partial(self.choose_viewpoint,"wide"),
            (KEYDOWN,K_PAUSE):self.toggle_pause,
            (KEYDOWN,K_a):self.toggle_auto,
            (KEYDOWN,K_f):pygame.display.toggle_fullscreen,
            }

    def __del__(self):
        lighting.release_light(self.marker_light)

    def event(self,event):
        t = event.type
        if t in (KEYUP,KEYDOWN):
            fn = self.keymap.get((t,event.key))
            if fn: fn()
        elif event.type == QUIT:
            self.quit = True
            self.racing = False

    def step(self):
        self.step_contents()
        if self.racing and not self.paused:
            if not self.won:
                self.drive_player()
            pcar = self.player_car
            ranks = [(pcar.finished_rank,-pcar.lap_number, len(pcar.lap),pcar)]
            for ccar in self.computer_cars:
                ranks.append((ccar.finished_rank,-ccar.lap_number, len(ccar.lap),ccar))
                if not ccar.running: continue
                self.drive_computer(ccar)
            ranks.sort()
            where_is_player = [i for (i,(r0,r1,r2,c)) in enumerate(ranks) if c is pcar][0]
            self.lap_counter[1] = ["1st","2nd","3rd","4th","5th"][where_is_player]
        else: # waiting for race to start
            self.clock.tick()
            self.one_second = self.clock.get_fps() # calibrate times
            if camera.ready() and not (self.quit or self.won):
                self.racing = True
                self.race_start_time = time.time()
                self.lap_counter[0] = "Lap 1/%d" % self.track.laps
                self.player_car.keyturn = 1.0
                for ccar in self.computer_cars:
                    ccar.keyturn = 1.0

    def drive_player(self):
        track = self.track
        pcar = self.player_car
        if self.auto_steering:
            pcar.auto_steer(track)
        else:
            pcar.brake(self.braking)
            steering = self.steering_right - self.steering_left
            pcar.turn(steering)
        wind_vec = self.twister.wind_vec(*pcar.pos)
        self.wind_pointer.point_along(wind_vec.x,wind_vec.y,pcar.angle)
        pcar.drive(track,wind_vec)
        howl = min(wind_vec.length()*5,0.5)
        pygame.mixer.music.set_volume(howl)
        if pcar.gone:
            self.spawn_at_checkpoint(pcar)
        else:
            self.update_view()
            x,y,z = pcar.pos
            tile = track.tile_at(x,y)
            if pcar.lap and pcar.lap[0] == tile:
                pcar.last_checkpoint = (x,y,pcar.angle)
                del pcar.lap[:1]
                if pcar.lap:
                    tx,ty = pcar.lap[0]
                    self.set_marker(tx,ty)
                else:
                    self.end_lap(pcar,True)
            mx,my,mz = pcar.marker_coords
            self.course_pointer.point_along(mx-x,my-y,pcar.angle)
        t = time.time() - self.race_start_time
        seconds = fmod(t,60.0)
        minutes = int(t - seconds) // 60
        self.race_timer[0] = "%02d:%05.2f" % (minutes,seconds)

    def drive_computer(self,ccar):
        track = self.track
        ccar.auto_steer(track)
        ccar.drive
        wind_vec = self.twister.wind_vec(*ccar.pos)
        ccar.drive(track,wind_vec)
        if ccar.gone:
            self.spawn_at_checkpoint(ccar)
        else:
            x,y,z = ccar.pos
            tile = track.tile_at(x,y)
            if ccar.lap and ccar.lap[0] == tile:
                ccar.last_checkpoint = (x,y,ccar.angle)
                del ccar.lap[:1]
                if not ccar.lap:
                    self.end_lap(ccar,False)
        
    def escape_key(self):
        self.quit = True
        self.racing = False
        pygame.mixer.music.fadeout(3000)
        four_seconds = int(4*self.one_second)
        camera.look_from((0,-100,200),four_seconds)

    def toggle_pause(self):
        self.paused = not self.paused
    def toggle_auto(self):
        self.auto_steering = not self.auto_steering
        
    def steer(self,left,on):
        self.auto_steering = False
        if left:
            self.steering_left = on
        else:
            self.steering_right = on

    def brake(self,on):
        self.braking = on
        self.auto_steering = False
    
    def choose_viewpoint(self,vptype):
        self.viewpoint = vptype

    def update_view(self):
        pcar = self.player_car
        x,y,z = pcar.pos
        vp = self.viewpoint
        if vp == "car":
            camera.look_at((x,y,z+3.0))
            camera.look_from_spherical(0.0,180.0+pcar.angle,15.0,25)
            camera.choose_up((0,0,1))
        elif vp == "bird":
            camera.look_at((x,y,z))
            camera.look_from((x,y,100))
            a = radians(pcar.angle)
            camera.choose_up((cos(a),sin(a),0))
        elif vp == "inside":
            camera.look_at((x,y,z+1.0))
            camera.look_from_spherical(0.0,180.0+pcar.angle,0.1)
            camera.choose_up((0,0,1))
        elif vp == "wide":
            camera.look_at((x,y,z+5.0))
            camera.look_from_spherical(0.0,180.0+pcar.angle,15.0,30)
            camera.choose_up((0,0,1))

    def expired(self):
        return (self.won or self.quit) and camera.ready()

    def build_parts(self):
        track = self.track = self.options["track"]
        if track.twister_texture:
            stylesheet.load({"Tornado":{"texture":track.twister_texture}})
        if self.options["slow"]:
            racetrack.NOTEXTURES = True
            stylesheet.load({"Tornado":{"texture":None,"colour":(0.5,0.5,0.5,0.5)}})
        scape = racetrack.Landscape('ground',track)
        matname = self.options["material"]
        # Cars
        pcar = self.player_car = self.options["car"]
        pcar.angle = 0.0
        pcar.finished_rank = 100 # not finished yet
        pcar.finish_time = None
        pcar.keyturn = -5.0
        track.obstacles.append(pcar)
        self.spawn_at_start(pcar,int(4 * self.one_second))
        self.windup_sound = pygame.mixer.Sound(data.filepath("sound/windup.ogg"))
        self.windup_sound.set_volume(0.4)
        self.windup_sound.play()
        pygame.mixer.music.play(-1)
        self.player_car.lap = track.course[:]
        px,py,pz = self.player_car.pos
        computer_cars = self.computer_cars = []
        mat_list = random.sample([m for m in cars.car_materials if m != matname],4)
        for i in range(5):
            if i == 2: continue # player is in lane 2
            name,klass = random.choice(cars.all_cars)
            ccar = klass('CPU ' + ("12.34"[i]))
            ccar._style["override-mtl"] = mat_list.pop()
            cy = py + 4.0 - 2.0*i
            cz = track.z_at(px,cy)
            ccar.pos = (px,cy,cz)
            ccar.lap = track.course[:]
            ccar.finished_rank = 100 # not finished yet
            ccar.finish_time = None
            ccar.keyturn = -5.0
            track.obstacles.append(ccar)
            computer_cars.append(ccar)
            
        # Weather
        skyball=sky.SkyBall('sky')
        self.twister=tornado.Tornado('twister',track)
        # Course marker
        self.marker = objpart.ObjPart('marker')
        # ViewGroup
        vgroup = viewgroup.ViewGroup('view',
            [skyball,scape,self.twister,self.player_car,self.marker]
            +computer_cars,
            geom=dict(perspective_angle=30.0))
        self.append(vgroup)
        L = self.marker_light
        lighting.light_colour(L,(0.8,0.8,0.8,1))
        tx,ty = track.course[0]
        self.set_marker(tx,ty)
        # Starting post
        start = objpart.ObjPart("start")
        x,y = track.tile_xy(tx,ty)
        x += track.tilewidth
        y += track.tilewidth * 0.5
        z = track.z_at(x,y) - 0.5
        start.pos = (x,y,z)
        start.angle = -90
        vgroup.append(start)
        # Furniture
        for fname,tx,ty,angle in track.furniture:
            fx,fy = track.tile_xy(tx,ty)
            thing = objpart.ObjPart('',style={'obj-filename':fname},
                                    geom=dict(pos=(fx,fy,track.z_at(fx,fy)),
                                              angle=angle))
            vgroup.append(thing)
        # HUD
        cp = self.course_pointer = PointerDisplay("to_marker")
        wp = self.wind_pointer = PointerDisplay("wind_dir")
        wp.pos = (-1.0,-0.5,0)
        rtimer = self.race_timer = textpanel.TextPanel("rtimer",["00:00.00"])
        rtimer.pos = (-0.8,0.9,0)
        lapcount = self.lap_counter = textpanel.TextPanel("lapcount",
            ["Lap 0/%d" % track.laps,""])
        lapcount.pos = (0.8,0.8,0)
        geom = standard_ortho_geom()
        hud = viewgroup.OrthoGroup('hud',[cp,wp,rtimer,lapcount],geom=geom)
        self.append(hud)


    def set_marker(self,tx,ty,on=True):
        L = self.marker_light
        lighting.light_switch(L,on)
        if on:
            x,y = self.track.tile_xy(tx+0.5,ty+0.5)
            z = self.track.z_at(x,y)
            lighting.light_position(L,(x,y,z+2,1))
            self.marker.pos = x,y,z+2
            self.player_car.marker_coords = (x,y,z)

    def spawn_at_start(self,pcar,steps=240):
        track = self.track
        startx,starty = track.tile_xy(*track.start_tile())
        # centre on start tile
        startx += track.tilewidth / 2.0
        starty += track.tilewidth / 2.0
        startz = track.z_at(startx,starty)
        pcar.pos = (startx,starty,startz)
        pcar.park()
        if steps:
            camera.look_at((startx,starty,startz)) # look at player
            camera.look_from_spherical(80.0,180.0,100.0,0)
            camera.look_from_spherical(10.0,180.0,40.0,steps)

    def spawn_at_checkpoint(self,pcar):
        try:
            x,y,a = pcar.last_checkpoint
            z = self.track.z_at(x,y)
            pcar.park()
            pcar.pos = (x,y,z)
            pcar.angle = a
        except AttributeError:
            self.spawn_at_start(pcar,0)

    def end_lap(self,pcar,isplayer=False):
        track = self.track
        lap_number = pcar.lap_number
        if lap_number >= track.laps:
            self.cars_finished += 1
            pcar.finished_rank = self.cars_finished
            pcar.finish_time = time.time() - self.race_start_time
            if isplayer:
                self.win()
            else:
                pcar.running = False
        else:
            pcar.lap_number += 1
            pcar.lap = track.course[1:] # not start square again
            if isplayer:
                tx,ty = pcar.lap[0]
                self.set_marker(tx,ty)
                self.lap_counter[0] = "Lap %d/%d" % (pcar.lap_number,track.laps)

    def win(self):
        self.won = True
        pcar = self.player_car
        x,y,z = pcar.pos
        camera.look_at((x,y,z))
        camera.look_from_spherical(10.0,-180.0,10.0)
        camera.look_from_spherical(10.0,180.0,50.0,int(10*self.one_second))
        # TODO play victory music...
        self.twister._expired = True
        pygame.mixer.music.fadeout(3000)
        self["sky"]._style_classes = ["clear"]
        self["sky"].restyle()

    def fill_rankings(self):
        rankings = self.options["rankings"] = []
        pcar = self.player_car
        nicknames = {("Crocodile","Chocolate"):"Chocodile"}
        pname = nicknames.get((pcar.__class__.__name__,self.options["material"]),
                              "Player 1")
        pcar.track = None
        rankings.append((pcar.finished_rank, pcar.finish_time,pname,pcar))
        for ccar in self.computer_cars:
            ccar.track = None
            rankings.append((ccar.finished_rank,ccar.finish_time,ccar._name,ccar))
        rankings.sort()

    def inexplicable_lights_hack(self):
        lighting.release_light(self.marker_light)
        lighting.release_light(self["view"].light)
        self.contents = []

class CarLine(part.Group):
    """ for list of cars on the rankings screen """
    _default_geom = dict(pos=(0,0,0),angle=0.0,scale=0.2)
    def __init__(self,name,contents,**kw):
        super(CarLine,self).__init__(name,contents,**kw)
    def setup_geom(self):
        """Move to pos, turn by angle and scale by scale"""
        super(CarLine,self).setup_geom()
        scale=self._geom.get('scale',1)
        if scale and scale != 1:
            glEnable(GL_NORMALIZE)
            glScale(scale,scale,scale)
        glRotate(-85.0,1.0,0,0) # side-on view, slightly above
    def setup_style(self):
        glPushAttrib(GL_LIGHTING_BIT)
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT,(0.8,0.8,0.8,1.0))
    def setdown_style(self):
        glPopAttrib()

class RankingsScreen(part.Group):
    def __init__(self,options,**kw):
        super(RankingsScreen,self).__init__('rankings',contents=[],**kw)
        back = backdrop.Backdrop('back')
        self.append(back)        
        banner = Banner('banner',geom=dict(left=-1.3,bottom=0.5,width=1.0,height=0.5))
        ranktext = []
        ranked_cars = []
        carx = 5.0
        for rank,rtime,name,car in options["rankings"]:
            car.pos = (carx,0,0)
            car.angle = -30.0
            ranked_cars.append(car)
            carx -= 2.5
            if rtime is not None:
                minutes = int(rtime // 60)
                seconds = rtime - (minutes*60)
                ranktext.append("%d %-10.10s %02d:%05.2f" % (rank,name,minutes,seconds))
            else:
                ranktext.append("- %-10.10s No finish" % name)
        rankings = textpanel.TextPanel('panel',ranktext)
        
        geom = standard_ortho_geom()
        cargroup = CarLine("cars",ranked_cars)
        cargroup.pos = (0,-0.9,0)
        ogroup = viewgroup.OrthoGroup('ortho',[banner,rankings,cargroup],geom=geom)
        self.append(ogroup)

    def event(self, event):
        choice = None
        if event.type in (QUIT,KEYDOWN):
            self._expired = True
    
def run_title_screen():
    stylesheet.load(default_stylesheet)
    stylesheet.load(title_stylesheet)
    t = TitleScreen()
    mainloop.until_expired(t)
    return t.choice

def run_option_screen(options):
    o = OptionsScreen(options)
    mainloop.until_expired(o)

def choose_car(options):
    stylesheet.load(default_stylesheet)
    stylesheet.load(cars_stylesheet)
    c = CarScreen(options)
    mainloop.until_expired(c)
    return "Back" if c.quit else None

def choose_track(options):
    stylesheet.load(default_stylesheet)
    stylesheet.load(tracks_stylesheet)
    t = TrackScreen(options)
    mainloop.until_expired(t)
    if t.quit:
        return "Back"
    else:
        track = options["track"]
        if track:
            track.start_compile()

def choose_colour(options):
    stylesheet.load(default_stylesheet)
    c = ColourScreen(options)
    mainloop.until_expired(c)
    return "Back" if c.quit else None

def run_race(options):
    stylesheet.load(default_stylesheet)
    stylesheet.load(race_stylesheet)
    r = RaceScreen(options)
    mainloop.until_expired(r,showrate=True)
    r.fill_rankings()
    r.inexplicable_lights_hack()

def show_rankings(options):
    stylesheet.load(default_stylesheet)
    r = RankingsScreen(options)
    mainloop.until_expired(r)
    
