"""
    racetrack.py   RaceTrack objects

   Pyweek 5 Scavengers team entry  Sep 2007

   GPL v3 or later
"""
from __future__ import division
from array import array
from threading import Thread
import logging
from math import fmod
import struct

import pygame
from glob import glob
from OpenGL.GL import *
from tdlib import part, texture, picking, vec, floorset

import data
import objpart
import pdb

NOTEXTURES=False
NegativeHuge = -100
MapScale = 2

class HeightField(object):
    """A height field loaded from an indexed PNG file"""
    def __init__(self, pngname, topvalue=16.0, offset=0.0):
        img = pygame.image.load(data.filepath(pngname))
        self.width, self.length = img.get_size()
        self.array = array('f',(ord(c) * topvalue / 255.0 + offset
                                for c in pygame.image.tostring(img,"P",True)))
    def z_at(self,x,y):
        x = int(x)
        y = int(y)
        if 0 <= x < self.width and 0 <= y < self.length:
            i = self.width * y + x
            return self.array[i]
        else:
            return NegativeHuge

class TrackCompilerThread(Thread):
    """ Compile the height-field of a RaceTrack into Vertex Arrays
        for glDrawElements """
    def __init__(self,track,*kw):
        Thread.__init__(self,*kw)
        self.track = track
    def run(self):
        track = self.track
        W = int((track.topo.width-1) / track.tiles_wide)
        Nvertices = (W+1)*(W+1)
        Nindices = W*W*6
        indices = array('i',range(Nindices))
        vertices = array('f',[0.0] * Nvertices * 3)
        normals = array('f',[0.0] * Nvertices * 3)
        track.vertex_index_data = ""
        # Build indices into vertex array (same for all tiles)
        cell_index_order = [0,W+2,W+1,0,1,W+2]
        i0 = 0
        j0 = 0
        for cy in range(W):
            for cx in range(W):
                indices[j0:j0+6] = array('i',(i+i0 for i in cell_index_order))
                j0 += 6
                i0 += 1
            i0 += 1
        track.vertex_index_data = indices.tostring()
        track.num_vertex_indices = Nindices
        # For each tile, build a vertex and a normals array
        track.vertex_data = {}
        track.normals_data = {}
        for ty in range(track.tiles_long):
            for tx in range(track.tiles_wide):
                iv = 0
                x0, y0 = track.tile_xy(tx,ty)
                for iy in range(W+1):
                    y = y0 + iy*MapScale
                    for ix in range(W+1):
                        x = x0 + ix*MapScale
                        z = track.z_at(x,y)
                        vertices[iv] = x
                        vertices[iv+1] = y
                        vertices[iv+2] = z
                        iv += 3
                track.vertex_data[tx,ty] = vertices.tostring()
                for i in range(0,Nindices,3):
                    js = indices[i:i+3]
                    vs = [vertices[j*3:j*3+3] for j in js]
                    ns = array('f',vec.plane_normal(*vs))
                    for j in js:
                        normals[j*3:j*3+3] = ns
                track.normals_data[tx,ty] = normals.tostring()
        # now can use glVertexPointer, glNormalPointer and glDrawElements
        logging.info("track compiler thread done...")
        return                         
                
class RaceTrack(object):
    """The data for a racetrack, consisting of:
        a height field
        a set of tile textures for drawing sections of the ground
        a list of tiles that must be crossed for one lap of the course
        a bunch of scenery/furniture
    """
    def __init__(self, trackfilename):
        self.textures = {}
        tlegend = {}
        tracklines = data.load(trackfilename,"r")
        self.tilewidth = None
        self.course = []
        self.title = "Untitled"
        self.compiler_thread = None
        self.obstacles = []
        self.furniture = []
        self.twister_texture = None
        for line in tracklines:
            tokens = line.split()
            if not tokens or line.startswith('#'):
                continue
            key = tokens[0]
            if key == 'heights':
                self.topo_file = tokens[1]
                self.topo = None
                self.load_heights()
            elif key == 'texture':
                filename = data.filepath(tokens[2])
                c = tokens[1]
                tlegend[c] = texture.get_texture(filename)
            elif key == 'map':
                if self.tilewidth is None:
                    self.tiles_wide = len(tokens[1])
                    self.tilewidth = self.width // self.tiles_wide
                    tile_row = int(self.length // self.tilewidth)
                    self.tiles_long = tile_row
                tile_row -= 1
                if tile_row < 0: # too many rows
                    continue
                for tile_col, c in enumerate(tokens[1]):
                    self.textures[tile_col, tile_row] = tlegend.get(c)
            elif key == 'course':
                for i in range(1,len(tokens)):
                    self.course.append(tuple(map(int,tokens[i].split(','))))
            elif key == 'title':
                self.title = ' '.join(tokens[1:])
            elif key == 'laps':
                self.laps = int(tokens[1])
            elif key == 'obstacle':
                fname = tokens[1]
                x,y,angle,radius,height = map(float,tokens[2:7])
                self.furniture.append((fname,x,y,angle))
                self.add_obstacle(x,y,radius,height)
            elif key == 'twister':
                self.twister_texture = tokens[1]

    def load_heights(self):
        self.topo = HeightField(self.topo_file)
        self.width = (self.topo.width - 1) * MapScale
        self.length = (self.topo.length - 1) * MapScale
        self.xoffset = self.width / 2.0
        self.yoffset = self.length / 2.0
        
    def start_compile(self):
        self.load_heights()
        if self.compiler_thread:
            return # already running, I assume
        self.compiler_thread = TrackCompilerThread(self)
        self.compiler_thread.start()

    def z_at(self,x,y):
        if self.topo:
            # convert to height-field coords
            hfx = (x+self.xoffset) / MapScale
            hfy = (y+self.yoffset) / MapScale
            return self.topo.z_at(hfx,hfy)
        else:
            return 0.0

    def tile_at(self,x,y):
        return (int(x + self.xoffset) // self.tilewidth,
                int(y + self.yoffset) // self.tilewidth)

    def tile_xy(self,tx,ty):
        return (tx*self.tilewidth - self.xoffset,
                ty*self.tilewidth - self.yoffset)

    def start_tile(self):
        return self.course[0] if self.course else (0,0)

    def normal_at(self,x,y):
        """Use normals pre-calculated by compiler thread"""
        if self.compiler_thread:
            return (0,0,1) # not ready yet
        tx,ty = self.tile_at(x,y)
        try:
            ndata = self.normals_data[tx,ty]
        except KeyError:
            return (0,0,1)
        tw = self.tilewidth
        hfx = fmod(x+self.xoffset,tw) / MapScale
        hfy = fmod(y+self.yoffset,tw) / MapScale
        # integer and fractional parts
        qx,qy = int(hfx), int(hfy)
        rx = hfx - qx
        ry = hfy - qy
        # offset within vertex index data
        W = int(tw / MapScale)
        i = (qx + qy*W*2 + (rx>ry)) * 3 * 4
        packed_index = self.vertex_index_data[i:i+4]
        j = struct.unpack('i',packed_index)[0] * 12
        normal = struct.unpack('fff',ndata[j:j+12])
        return normal

    def add_obstacle(self,tx,ty,radius,height):
        x,y = self.tile_xy(tx,ty)
        z = self.z_at(x,y)
        self.obstacles.append(floorset.CylinderRegion((x,y,z),height,radius))

    def first_collision(self,notme,x,y,z,r,h=1.0):
        for o in self.obstacles:
            if o is notme:
                continue
            zs = o.overlaps((x,y,z),h,r)
            if zs: return o
        return None

class Landscape(part.Part):
    _default_style={'colour':(1,1,1,1), 'shiny':20.0}
    def __init__(self,name,track,**kw):
        self.dlbase = 0
        self.ndls = 0
        super(Landscape,self).__init__(name,**kw)
        self.track = track
        self.ndls = int(track.tiles_wide * track.tiles_long)
        self.dlbase = glGenLists(self.ndls)
        
    def __del__(self):
        glDeleteLists(self.dlbase,self.ndls)

    def prepare(self):
        track = self.track
        if track.compiler_thread:
            print "waiting for track compiler thread"
            track.compiler_thread.join()
            track.compiler_thread = None
        self.colour = self.getstyle('colour',(1,1,1,1))
        self.shiny = self.getstyle('shiny',70.0)
        # Render vertexarray data into display lists
        index_data = track.vertex_index_data
        num_indices = track.num_vertex_indices
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_NORMAL_ARRAY)
        tw = track.tiles_wide
        tl = track.tiles_long
        for tx in range(0,tw):
            for ty in range(0,tl):
                dl = ty*tw + tx + self.dlbase
                glNewList(dl,GL_COMPILE)
                tex = track.textures.get((tx,ty))
                if tex and not NOTEXTURES:
                    tex.bind()
                glNormalPointer(GL_FLOAT,12,track.normals_data[tx,ty])
                glVertexPointer(3,GL_FLOAT,12,track.vertex_data[tx,ty])
                glDrawElements(GL_TRIANGLES,num_indices,GL_UNSIGNED_INT,
                               index_data)
                glEndList()
        glDisableClientState(GL_VERTEX_ARRAY)
        glDisableClientState(GL_NORMAL_ARRAY)
            
    def setup_style(self):
        if NOTEXTURES:
            glDisable(GL_TEXTURE_2D)
        else:
            glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR)
            glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR)
            glTexGenfv(GL_S,GL_OBJECT_PLANE,(1.0 / self.track.tilewidth,0.0,0.0,0.0))
            glTexGenfv(GL_T,GL_OBJECT_PLANE,(0.0, 1.0 / self.track.tilewidth,0.0,0.0))
            glEnable(GL_TEXTURE_GEN_S)
            glEnable(GL_TEXTURE_GEN_T)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,GL_REPEAT)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,GL_REPEAT)
        glColor4fv(self.colour)
        glMaterialfv(GL_FRONT,GL_AMBIENT,(0.5,0.5,0.5,0.5))
        glMaterialfv(GL_FRONT,GL_DIFFUSE,self.colour)
        glMaterialfv(GL_FRONT,GL_SPECULAR,(1,1,1,1))
        glMaterialf(GL_FRONT,GL_SHININESS,self.shiny)
        
    def setdown_style(self):
        glDisable(GL_TEXTURE_GEN_S)
        glDisable(GL_TEXTURE_GEN_T)
        
    def render(self,mode):
        if mode != 'OPAQUE':
            return
        track = self.track
        tw = track.tiles_wide
        tl = track.tiles_long
        for tx in range(0,tw):
            for ty in range(0,tl):
                dl = ty*tw + tx + self.dlbase
                glCallList(dl)
        
class LandscapeThumb(part.Part):
    """Thumbnail of the racetrack, just texture, not heights"""
    
    def __init__(self,name,track,**kw):
        super(LandscapeThumb,self).__init__(name,**kw)
        self.dl = glGenLists(1)
        self.track = track

    def __del__(self):
        glDeleteLists(self.dl,1)
            
    def prepare(self):
        track = self.track
        w = track.tilewidth
        tw = track.tiles_wide
        th = track.tiles_long
        qw = 2.0 / tw
        glNewList(self.dl,GL_COMPILE)
        for tx in range(0,tw):
            x0 = 2.0 / tw * tx - 1.0
            z0 = 0.0
            for ty in range(0,th):
                tex = track.textures.get((tx,ty))
                if tex:
                    tex.bind()
                glBegin(GL_QUADS)
                for dx,dz,s,t in [(0,0,0,0),(qw,0,1,0),(qw,qw,1,1),(0,qw,0,1)]:
                    glTexCoord2f(s,t)
                    glVertex3f(x0+dx,0,z0+dz)
                glEnd()
                z0 += qw
        glEndList()

    def setup_style(self):
        glColor4f(1,1,1,1)
        glEnable(GL_TEXTURE_2D)
        glMaterialfv(GL_FRONT,GL_AMBIENT,(1,1,1,1))
        glMaterialfv(GL_FRONT,GL_DIFFUSE,(1,1,1,1))
        glMaterialfv(GL_FRONT,GL_SPECULAR,(1,1,1,1))
        glMaterialf(GL_FRONT,GL_SHININESS,0.0)        
        
    def render(self,mode):
        glCallList(self.dl)        
     
def all_tracks():
    return (RaceTrack(tfile)
            for tfile in glob(data.filepath("tracks/*.track")))
