#! /usr/bin/env python
"""Generic routines that might be useful for animation by interpolating
   values.

    Copyright 2006 Peter Harris
    Released under the terms of the GNU General Public License
    (see www.gnu.org for details)
"""
from __future__ import division

def interpolator(xfrom,xto,steps):
    """Interpolate between two values, which should be both a number, a tuple of numbers
    or a dict with numeric values"""
    if type(xfrom) is tuple and type(xto) is tuple:
        return vec_interpolator(xfrom,xto,steps)
    elif type(xfrom) is dict and type(xto) is dict:
        return dict_interpolator(xfrom,xto,steps)
    else:
        return num_interpolator(xfrom,xto,steps)
        
def vec_interpolator(vfrom,vto,steps):
    """An iterator that interpolates between one vector of values to another
    in a given number of steps"""
    components = range(len(vfrom))
    current = list(vfrom)
    steps = int(steps)
    if steps:
        deltas = [(vto[i] - vfrom[i])/steps
                  for i in components]
    while steps > 0:
        for i in components:
            current[i] += deltas[i]
        steps -= 1
        yield tuple(current)
    yield tuple(vto)
        

def num_interpolator(nfrom, nto, steps):
    """An iterator that interpolates between one number and another in a given number of
    steps"""
    current = nfrom
    steps = int(steps)
    if steps:
        delta = (nto - nfrom)/steps
    while steps > 0:
        current += delta
        steps -= 1
        yield current
    yield nto
    
    
def dict_interpolator(dfrom,dto,steps):
    """An interpolator for dicts instead of tuples"""
    components = dfrom.keys()
    current = dto.copy()
    last = dto.copy()
    current.update(dfrom)
    steps = int(steps)
    if steps:
        deltas = [(dto.get(k,dfrom[k]) - dfrom[k])/steps
                  for k in components]
    while steps > 0:
        for i, k in enumerate(components):
            current[k] += deltas[i]
        steps -= 1
        yield current.copy()
    yield last

def oscillator(xfrom,xto,steps):
    """Go from one value to another and back indefinitely.
    NB: this generator is infinite. Be careful.
    """
    while steps:    # just give up if steps==0 ; makes no sense
        for x in interpolator(xfrom,xto,steps):
            yield x
        for x in interpolator(xto,xfrom,steps):
            yield x

def sequencer(xlist,steplist):
    """Interpolate between a series of values in succession"""
    xiter = iter(xlist)
    stepiter = iter(steplist)
    xfrom = xiter.next()
    steps = stepiter.next()
    for xto in xiter:
        for x in interpolator(xfrom,xto,steps):
            yield x
        xfrom = xto
        try:
            steps = stepiter.next()
        except StopIteration:
            pass    # keep same number of steps for the rest of them
    
class Animator(object):
    """A collection of numeric values which may be static or changing"""
    def __init__(self, **initvalues):
        self.current={}
        self.current.update(initvalues)
        self.interpolators={}
    def step(self):
        """Get next value of each interpolator, change current values"""
        for k,it in list(self.interpolators.items()):
            try:
                self.current[k] = it.next()
            except StopIteration:
                del self.interpolators[k]
    def finished(self,name=None):
        """Whether a particular value has finished changing,
        or all finished if no name is specified"""
        if name is None:
            return not(self.interpolators)
        else:
            return name not in self.interpolators
    def items(self):
        """current value items"""
        return self.current.items()
    def __getitem__(self,name):
        """Get a current value - defaults to zero - never a KeyError"""
        return self.current.get(name,0.0)
    def get(self,name,default=0.0):
        """Get a current value, with specified default"""
        return self.current.get(name,default)
    def __delitem__(self,name):
        """Delete a value and its interpolator if there is one"""
        try:
            del self.interpolators[name]
        except KeyError:
            pass
        try:
            del self.current[name]
        except KeyError:
            pass
    def __setitem__(self,name,value):
        """Set a value, delete any interpolator that may be present"""
        self.current[name] = value
        try:
            del self.interpolators[name]
        except KeyError:
            pass
    def copy(self):
        """Return a shallow copy of the Animator's current state.
        There's no point in copying the interpolators - you can't
        in general copy a generator."""
        return Animator(**self.current)
    def update(self,other):
        """Update by setting current values from another dict-like object."""
        for k,x in other.items:
            self[k] = x
    def change(self,name,xto,steps=0):
        """Begin changing a value"""
        if steps:
            self.interpolators[name] = interpolator(self.current.get(name,xto),
                                                    xto,steps)
        else:
            self[name] = xto
    def oscillate(self,name,xfrom,xto,steps=1):
        """Begin oscillating a value
        NB: it will NEVER be finished, until replaced with change() method.
        """
        self.interpolators[name] = oscillator(xfrom,xto,steps)
    def sequence(self,name,xlist,steplist):
        """Begin a more complex sequence of changes"""
        self.interpolators[name] = sequencer(xlist,steplist)

class Mutator(Animator):
    """An animator that works on another dict-like object"""
    def __init__(self,mutand):
        self.current = mutand
        self.interpolators = {}
