""" commander.py - pluggable CommandInterpreter class

  CommandInterpreter stores the commands it recognises in a
  dictionary keyed by command name.
  
  A command name is either a sequence of alphanumeric chars
  or a single punctuation character.
  
  Under each command name are as many variants as you like,
  consisting of:
  ( pattern - a regular expression to match the rest of the line
        after the command
    function - a function that will be called with keyword args
        taken from the named groups matching the pattern. The command
        itself will be inserted as keyword argument 'command'.
    help - a string giving the usage of this variant of the command.
  )
  
  You add commands by calling the plug() method:
    plug(command,pattern,function,help)

  If there is a commanchain attribute, unknown commands are passed
  to it, assuming it is CommandInterpreter.
  
  help can be got from commands as a list of help strings matching
  an initial substring provided.
  
  autoplug(object,prefix="cmd_") scans object for callable members
  whose names begin with the prefix, and plugs them in, using their
  __doc__ string to get the command name and syntax pattern
  
  
  Copyright 2005 Peter Harris
  Released under the terms of the GNU General Public License
"""
import re
import warnings

def safe_eval(expr,variables):
    """evaluate expression in a semi-safe environment"""
    blocked ={'file':None, 'open':None, 'eval':None, 'execfile':None,
              '__builtins__':None}
    blocked.update(variables)
    return eval(expr,blocked)

# regular expression used to pick command off front of the command line
rx_commandsplit = re.compile("""
    \s*  # initial whitespace
    (?P<command>        # The command group is:
        [!-/:-@[-`{-~]  #   a single punctuation character
    |   [0-9a-zA-Z]+    #   or some alphanumerics
    )
    \s*                 # whitespace between command and tail stripped
    (?P<tail>.*)        # everything else is the tail
    """, re.VERBOSE)
    
def splitcommand(cmdline):
    """return (command,rest of line)"""
    match = rx_commandsplit.match(cmdline)
    if match:
        return match.group('command','tail')
    else:
        return ('','')

class NotRecognised(Exception):
    pass

class CommandInterpreter(object):
    """An interpreter for simple commands with syntax that can be
    parsed using a single regular expression"""
    
    def __init__(self,*args,**kw):
        super(CommandInterpreter,self).__init__(*args,**kw)
        self.commanddict = {}
        self.output = None
        self.commandchain = None
        self.reflags = re.VERBOSE
        
    def plug(self,cmd,pattern,function,help):
        """Plug in a function to be called for commands matching a pattern"""
        if not help.strip().startswith(cmd):
            warnings.warn("Unhelpful help for command '%s':%s" % (cmd,help))
        rx = re.compile(pattern,self.reflags)
        try:
            self.commanddict[cmd].append((rx,function,help))
        except KeyError:
            self.commanddict[cmd] = [(rx,function,help)]
            
    def autoplug(self, obj=None, prefix='cmd_'):
        """Plug in methods with properly formatted doc strings as commands
        help text is between a line containing only <help> and a line
        containing only </help>,
        syntax pattern is likewise between <pattern> and </pattern> lines.
        Anything else is function documentation, left alone.        
        """
        if obj is None:
            obj = self
        for name in dir(obj):
            if not name.startswith(prefix):
                continue
            meth = getattr(obj,name)
            if not callable(meth) or not meth.__doc__:
                continue
                
            doclines = meth.__doc__.split('\n')
            helplines = []
            patternlines = []
            which = None
            for line in doclines:
                clean = line.strip()
                if clean == '<help>':
                    which = helplines
                elif clean in ('</help>','</pattern>'):
                    which = None
                elif clean == '<pattern>':
                    which = patternlines
                elif which != None:
                    which.append(line)
            helptext = '\n'.join(helplines)
            pattern = '\n'.join(patternlines)
            if helptext and pattern:
                cmd = splitcommand(helptext)[0]
                self.plug(cmd,pattern,meth,helptext)

    def default_help_command(self,helpword='help'):
        """Set up a command to do the default help behaviour"""
        self.plug(helpword,
            '(?P<command>.*)',
            self.gethelp,
            '%s\t: get help' % helpword)

    def gethelp(self,command=''):
        """get help strings for a command prefix"""
        cmd,tail = splitcommand(command)
        commandline = command.strip()   # take off whitespace
        variants = self.commanddict.get(cmd,())
        helplist = []
        for rx, function, help in variants:
            if help.startswith(commandline):
                helplist.append(help)
        if self.commandchain:
            helplist.extend(self.commandchain.gethelp(command=commandline))
        return helplist

    def execute(self, command):
        """Execute a command if it matches any patterns known"""
        matched = False
        result = None
        cmd,tail = splitcommand(command)
        commandline = command.strip()   # take off whitespace
        variants = self.commanddict.get(cmd,())
        for rx, function, help in variants:
            match = rx.match(tail)
            if match:
                result = function(**match.groupdict())
                matched = True
                self.respond(result)
                break   # and we're done
        if not matched:
            if self.commandchain:
                result = self.commandchain.execute(commandline)
                self.respond(result)
            else:
                print tail, 'does not match' #DEBUG
                raise NotRecognised(commandline)
        return result
    
    def respond(self,message):
        """Feed message to stream or list or whatever it is"""
        output = self.output
        if output is not None:
            if callable(output):
                output(message)
            elif hasattr(output, 'append'): # special case for lists
                output.append(message)
            else:
                print >> output, str(message)    # special case for streams
