""" FootPrint - an object that can tell whether it overlaps a circle
    in the x,y plane.
    FootPrints can be an arbitrarily rotated rectangle, or a circle,
    which are quite simple to calculate.
    They can also be general convex polygons, which are less
    straightforward.

    Usage:
    Supposing there are obstacles on a floor. Whether the hero can walk on
    a bit of floor depends on whether his footprint (for the sake of
    simplicity, a circle) overlaps any of the obstacles.

    Copyright 2004 Peter Harris
    Released under the terms of the GNU General Public License
    see www.gnu.org for details.
    
    OBSOLETE - code moved to floorset.py

"""
from math import sin, cos, radians
from tdlib.vec import Vec

class Circle(object):
    def __init__(self,x,y,r):
        self.x = x
        self.y = y
        self.r = r
    def overlaps(self,x,y,r=0):
        """Distance between centres is less than sum of r and own radius"""
        dx = x - self.x
        dy = y - self.y
        d2 = dx*dx + dy*dy
        r2 = (r+self.r)**2
        return d2 < r2

class RotatedRect(object):
    def __init__(self,x,y,width,length,angle=0):
        self.x = x
        self.y = y
        self.w = width / 2.0
        self.l = length / 2.0
        # unit vector parallel to one side of the rectangle
        self.u = (cos(radians(angle)), sin(radians(angle)))
        # unit vector normal to that one
        self.v = (-self.u[1],self.u[0])
    def overlaps(self,x,y,r=0):
        """Find distance from circle centre to lines through centre of the
        rectangle parallel to the sides."""
        # vector P is from centre of rectangle to centre of circle
        px = x - self.x
        py = y - self.y
        # By the magic of the vector dot product...
        ux,uy = self.u
        vx,vy = self.v
        du = abs(px * ux + py * uy)  # component of P in direction of u
        dv = abs(px * vx + py * vy)  # component of P in direction of v
        return (du - r) < self.w and (dv - r) < self.l

class Polygon(object):
    """A Polygon footprint uses some vector algebra to decide whether
    a point lies within it, and if not then whether it lies within
    a given distance of any part of it.
    
    Some conjectures (no proof offered)
      A if you follow the boundary of a convex polygon anticlockwise,
        any point outside the polygon will at some point be on your right.
        
      B if a disc whose centre is outside the polygon intersects any sides
        of it, it must intersect at least one of the sides for which the
        centre of the disc is "on the right" as above. The others are on the
        side of the polygon facing away.
        
      C if the centre of the disc is on the right of a side and it is further
        away from the line passing through that side than the disc's radius, it
        cannot intersect the polygon anywhere. (All other sides are "on the left"
        so no part of the disc can reach them.)
    """
    def __init__(self,x,y,points,angle=0):
        self.x = x
        self.y = y
        self.angle=angle
        self.sine = sin(radians(-angle))
        self.cosine = cos(radians(-angle))
        self.points = points    # must be convex, sorted anticlockwise
    
    def overlaps(self,x,y,r=0):
        """Decide if centre of disc (x,y) is outside the polygon.
        If so, see if it crosses anywhere."""
        # first rotate (x,y) back by the angle the polygon was rotated to
        # (which is cheaper than rotating all the vertices of the polygon)
        dx = (x - self.x)
        dy = (y - self.y)
        s,c = self.sine, self.cosine
        r2 = r * r  # cheaper to compare squared distance than do sqrts
        P = Vec(dx * c - dy * s, dx * s + dy * c)
        v0 = Vec(self.points[-1]).projXY()
        any_outside = False
        for vtuple in self.points:
            v1 = Vec(vtuple).projXY()
            u = v1 - v0 # vector along side of polygon
            uperp = Vec(u.y, -u.x) # perpendicular to u, away from polygon
            w = P - v0  # vector from v0 to P
            outside = (w.dot(uperp) > 0) # w in "same direction" as uperp 
            if outside: # Conjectures A, B
                any_outside = True
                wproj = (w.proj(uperp)) # length is dist from side
                d2 = wproj.dot(wproj)
                if d2 > r2:
                    return False    # Conjecture C
                behind_v0 = (u.dot(w) < 0)
                if behind_v0: # v0 is nearest point to P
                    if w.dot(w) < r2:  # and within radius
                        return True
                else:
                    v = P - v1
                    behind_v1 = (v.dot(-u) < 0)
                    if behind_v1:  #  v1 is nearest point to P
                        if v.dot(v) < r2:  # and within radius
                            return True
                    else: # P between v0 and v1, and we know it's near enough
                        return True
            v0 = v1 # proceeding anti-clockwise
        # all the way round without finding a crossing
        return not any_outside  # centre within polygon
