#! /usr/bin/env python
""" furniture.py
A miscellaneous collection of Part objects, for assembling indoor scenes

Copyright 2004 Peter Harris, released under the terms of the
GNU General Public License
"""
from __future__ import division

from OpenGL import GL, GLU
from tdlib import part, floorset, tnvpart, texture, animator
from tdlib.vec import plane_normal
    
class CarpetFloor(part.Part, floorset.Floor):
    """A sloped rectangle, to be included in the visible contents of a Floor"""
    DEFAULT_CONFIG = {'pos':(0,0,0), 'size':(1,1,1), 'slope':(0,0),
                    'colour':(1,1,1,1), 'tex-orient':0,
                    'repeat-x':1, 'repeat-y':1}
    STYLE_ATTRIBUTES = ('tex-orient','colour','repeat-x','repeat-y','texture')
    def __init__(self, name,**splus):
        floorset.SlopeFloor.__init__(self,name,**splus)
        self.restyle()
        tex = self.config.get('texture')
        if type(tex) is str:
            tex = texture.get_texture(tex)
        self.config['tex'] = tex
        tcoords = [(0,0),(1,0),(1,1),(0,1)]
        tex_orient = self.config.get('tex-orient',0)
        self.tcoords = tcoords[tex_orient:] + tcoords[:tex_orient]
        dzx, dzy = self.config['slope']
        self.normal = plane_normal((0,1,dzy),(0,0,0),(1,0,dzy))
        self.dlist = GL.glGenLists(1)
        self.build_dlist()
    def __del__(self):
        GL.glDeleteLists(self.dlist,1)
    def build_dlist(self):
        """Build a display list which will be used to draw the object"""
        GL.glNewList(self.dlist,GL.GL_COMPILE)
        GL.glPushAttrib(GL.GL_LIGHTING_BIT)
        GL.glEnable(GL.GL_COLOR_MATERIAL)
        glfb = GL.GL_FRONT_AND_BACK
        GL.glColorMaterial(glfb,GL.GL_AMBIENT_AND_DIFFUSE)
        t = GL.glTexCoord2f
        v = GL.glVertex3f
        config = self.config
        colour = config['colour']
        tex = config.get('tex')
        GL.glColor4f(*colour)
        if tex:
            tex.bind()
            GL.glEnable(GL.GL_TEXTURE_2D)
        else:
            GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glBegin(GL.GL_QUADS)
        GL.glNormal3f(*self.normal)
        width, length, dummy = config['size']
        repeat_x = config['repeat-x']
        repeat_y = config['repeat-y']
        dzx, dzy = config['slope']
        dx = width / repeat_x
        dy = length / repeat_y
        x = -(width/2)
        for xi in range(repeat_x):
            y = -(length/2)
            for yi in range(repeat_y):
                z = dzx * x + dzy * y
                vcoords = [(x,y,z),(x+dx,y,z+dzx*dx),
                           (x+dx,y+dy,z+dzx*dx+dzy*dy),(x,y+dy,z+dzy*dy)]
                for i in range(4):
                    t(*self.tcoords[i]); v(*vcoords[i])
                y += dy
            x+=dx
        GL.glEnd()
        GL.glPopAttrib()
        GL.glEndList()
    def render(self):
        """Call the display list"""
        GL.glCallList(self.dlist)
        
class RingFloor(part.PosPart, floorset.Floor):
    DEFAULT_CONFIG={'pos':(0,0,0), 'radius':1.0, 'inner_radius':0.0,
                    'colour':(1,1,1,1), 'slices':16, 'loops':4}
    STYLE_ATTRIBUTES = ('colour','texture','slices','loops')
    def __init__(self,name, **splus):
        floorset.Floor.__init__(self,name,**splus)
        self.dlist = GL.glGenLists(1)
        self.quadric = GLU.gluNewQuadric()
        self.restyle()
    def prepare(self):
        tex = self.config.get('texture')
        if type(tex) is str:
            tex = texture.get_texture(tex)
        self.config['tex'] = tex
        if tex:
            GLU.gluQuadricTexture(self.quadric,True)
        self.build_dlist()
    def __del__(self):
        GL.glDeleteLists(self.dlist,1)
        GLU.gluDeleteQuadric(self.quadric)
    def build_dlist(self):
        """Build a display list which will be used to draw the object"""
        GL.glNewList(self.dlist,GL.GL_COMPILE)
        q = self.quadric
        GL.glPushAttrib(GL.GL_LIGHTING_BIT)
        GL.glEnable(GL.GL_COLOR_MATERIAL)
        glfb = GL.GL_FRONT_AND_BACK
        GL.glColorMaterial(glfb,GL.GL_AMBIENT_AND_DIFFUSE)
        config = self.config
        colour = config['colour']
        tex = config.get('tex')
        radius = config['radius']
        inner_radius = config['inner_radius']
        slices = config['slices']
        loops = config['loops']
        GL.glColor4f(*colour)
        if tex:
            tex.bind()
            GL.glEnable(GL.GL_TEXTURE_2D)
        else:
            GL.glDisable(GL.GL_TEXTURE_2D)
        GLU.gluDisk(q,inner_radius, radius, slices, loops)
        GL.glPopAttrib()
        GL.glEndList()
    def render(self):
        """Call the display list"""
        GL.glCallList(self.dlist)
    
class Pit(part.PosPart, floorset.Obstacle):
    DEFAULT_CONFIG = {'pos':(0,0,0), 'radius':1.0, 'depth':1.0,
                     'colour':(1,1,1,1), 'slices':16, 'stacks':4}
    STYLE_ATTRIBUTES = ('colour','slices','stacks','texture')
    def __init__(self,name,**splus):
        self.name = name
        floorset.Obstacle.__init__(self,name,**splus)
        self.quadric = GLU.gluNewQuadric()
        self.restyle()
    def prepare(self):
        tex = self.config.get('texture')
        if type(tex) is str:
            tex = texture.get_texture(tex)
        self.config['tex'] = tex
        if tex:
            GLU.gluQuadricTexture(self.quadric,True)
        GLU.gluQuadricOrientation(self.quadric,GLU.GLU_INSIDE)
    def __del__(self):
        GLU.gluDeleteQuadric(self.quadric)
    def render(self):
        """Draw a pit."""
        config = self.config
        depth = config['depth']
        colour = config['colour']
        tex = config.get('tex',None)
        radius = config['radius']
        slices = config['slices']
        stacks = config['stacks']
        GL.glTranslate(0,0, -depth)    # drop to bottom of pit
        GL.glPushAttrib(GL.GL_LIGHTING_BIT)
        GL.glEnable(GL.GL_COLOR_MATERIAL)
        glfb = GL.GL_FRONT_AND_BACK
        GL.glColorMaterial(glfb,GL.GL_AMBIENT_AND_DIFFUSE)
        GL.glColor4f(*colour)
        if tex:
            tex.bind()
            GL.glEnable(GL.GL_TEXTURE_2D)
        else:
            GL.glDisable(GL.GL_TEXTURE_2D)
        GLU.gluCylinder(self.quadric, radius, radius, depth, slices, stacks)
        GL.glPopAttrib()

class Chest(tnvpart.TNVpart,floorset.Obstacle, floorset.Trigger):
    DEFAULT_CONFIG = {'pos':(0,0,0), 'angle':0.0, 'size':(1,0.5,0.5),
                     'tnv-file':'art/chest0.tnv', 'colour':(1,1,1,1),
                     'texture':'art/woodchest.png'}
    STYLE_ATTRIBUTES = ('tnv-file','colour','texture')
    def __init__(self,name,**splus):
        self.name = name
        self.restyle()
        self.config.update(splus)
        self.calculate_footprint()


    
    
