#! /usr/bin/env python
"""
A Pauser is a wrapper group around some objects that intercepts events.
While is not paused it passes all events except the P key, which makes it
be paused. It draws nothing except its contents.

While it is paused, it intercepts all events. The P key makes it unpause.
While paused, it draws its contents and then puts a 50% grey overlay in front
of them.

This file copyright Peter Harris 2005, released under the terms of the
GNU GPL. See www.gnu.org for details.

"""


from OpenGL import GL
from pygame.locals import KEYDOWN,QUIT, K_p

import part
import texturefont

class Pauser(part.Group):
    def __init__(self,*args,**kwd):
        super(Pauser,self).__init__(*args,**kwd)
        self.paused = False
    def setup_geom(self): pass  # no positioning
    def setdown_geom(self): pass
    def render(self,mode):
        """Overlay with a grey rect if paused
        """
        if mode == 'PICK' and self.paused:
            return  # no picking of pauser or anything behind it
        for d in self.contents:
            d.draw(mode)
        if self.paused:
            GL.glEnable(GL.GL_BLEND)
            GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
            GL.glDisable(GL.GL_LIGHTING)
            GL.glDisable(GL.GL_DEPTH_TEST)
            GL.glDisable(GL.GL_TEXTURE_2D)
            GL.glPushMatrix()   # MODELVIEW
            GL.glMatrixMode(GL.GL_PROJECTION)
            GL.glPushMatrix()   # PROJECTION
            GL.glLoadIdentity()
            GL.glOrtho(-1.0,1.0,-1.0,1.0,-1.0,1.0)
            GL.glMatrixMode(GL.GL_MODELVIEW)
            GL.glLoadIdentity()
            self.draw_overlay(mode)
            GL.glMatrixMode(GL.GL_PROJECTION)
            GL.glPopMatrix()    # PROJECTION
            GL.glMatrixMode(GL.GL_MODELVIEW)
            GL.glPopMatrix()    # MODELVIEW
    def draw_overlay(self,mode):
        if mode != 'TRANSPARENT':   # overlay is never opaque
            return
        GL.glColor4f(0.1,0.1,0.1,0.5)
        GL.glRectf(-1.0,-1.0,1.0,1.0)
    def event(self,event):
        """Pause/Continue if P pressed. Intercept all events while paused"""
        if event.type == KEYDOWN and event.key == K_p:
            self.paused = not self.paused
            return None
        elif not self.paused:
            return self.update_contents(event)
    def step(self):
        if not self.paused:
            self.step_contents()

class TextPauser(Pauser):
    _default_style = {'colour':(1,1,1,0.8),'pitch':10, 'font':None}
    _style_attributes = ('colour','pitch','font')
    def __init__(self,*args,**kwd):
        super(TextPauser,self).__init__(*args,**kwd)
        self.text = kwd.get('text','Paused')
        
    def prepare(self):
        style = self._style
        font = style.get('font')
        if font is None:
            self.font = texturefont.TextureFont()
        else:
            self.font = font
        self.width = self.font.width(self.text)
        self.scale = 1.0 / style['pitch']
        self.colour = style['colour']
        
    def draw_overlay(self,mode):
        GL.glColor4f(0.1,0.1,0.1,0.5)
        GL.glRectf(-1.0,-1.0,1.0,1.0)
        GL.glColor4f(*self.colour)
        GL.glScale(self.scale,self.scale,1.0)
        GL.glTranslate(-(self.width / 2.0), -0.5, 0.0)
        self.font.render(self.text)
