""" picking.py
    OpenGL selection helper functions
    
    Copyright May 2005 by Peter Harris
    
    Released under the terms of the GNU General Public License
    
"""

from OpenGL import GL
from OpenGL.GLU import gluPickMatrix

# API attributes
active = False          # whether picking is underway
pickRect = None
DEFAULT_BUFFER_SIZE = 1024

# Private attributes
_labelDict  = {}
_lastLabelNum = 0

def start(vpx,vpy,vpdx,vpdy,backwards=False):
    """Set up picking matrix for a region of the viewport, clear dictionary
    of objects associated with picking names (which are GL_INTs)
    
    Usage:
      Before setting up the projection call picking.start(x,y,dx,dy)
      where (x,y) is the centre of the target rectangle of the viewport.
      
    If you have already set up projection, need to call with backwards=True,
    which will get a copy of the projection matrix and multiply it back in
    after setting up the picking matrix.
    """
    global active, pickRect, _lastLabelNum
    if active:
        raise ValueError("picking already active")
    _labelDict.clear()
    _lastLabelNum = 0
    active = True
    pickRect = (vpx, vpy, vpdx, vpdy)
    GL.glSelectBuffer(DEFAULT_BUFFER_SIZE)
    GL.glRenderMode(GL.GL_SELECT)
    GL.glInitNames()
    GL.glPushName(0)
    if backwards:
        projection_matrix = GL.glGetFloatv(GL.GL_PROJECTION_MATRIX)
    project()
    if backwards:
        GL.glMultMatrixf(projection_matrix)
    
def project():
    """Set up picking projection if picking is active. This can be used
    from other modules to make sure the picking projection is set, in case
    something else may have messed up the projection matrix
    (e.g. a scene with parts in both Ortho and Perspective projections,
    all of which may need to be picked)
    Can be used when picking is not active, as a glLoadIdentity.
    *** Leaves matrix mode in GL_PROJECTION. ***
    """
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glLoadIdentity()
    if active:
        gluPickMatrix(*pickRect)

def label(ob):
    """Set a label for the object. Anything drawn up to the next label()
    or nolabel() that is picked will cause end() to include the object in
    its return value"""
    global _lastLabelNum
    _lastLabelNum += 1
    _labelDict[_lastLabelNum] = ob
    GL.glLoadName(_lastLabelNum)
    return _lastLabelNum
    
def nolabel():
    """End of drawing a labelled object, as far as picking is concerned"""
    GL.glLoadName(0)

def end():  # end --> [(minz,maxz,ob)]
    """End GL_SELECT mode and get a list of objects added by picking.label()
    that overlapped the target rectangle, together with their z-distance range
    """
    global active
    active = False
    pickNames = list(GL.glRenderMode(GL.GL_RENDER))
    GL.glMatrixMode(GL.GL_PROJECTION)
    GL.glLoadIdentity()
    GL.glMatrixMode(GL.GL_MODELVIEW)
    picklist = [(minz,maxz,_labelDict.get(ns[0])) for minz,maxz,ns in pickNames]
    picklist.sort()
    return picklist

    
