"""skeled_model

This module separates out most of the functionality of Skeled from its
user interface.

Copyright 8 May 2004 by Peter Harris
Released under the terms of the GNU General Public License
see www.gnu.org for details
"""
from __future__ import division

import cPickle as pickle
from math import sqrt, radians, sin, cos
from OpenGL.quaternion import quaternion
from OpenGL import GL

if 'set' not in dir(__builtins__):  # not 2.4 yet?
    from sets import Set as set     # use the Set class if builtin set not there

from tdlib import tnv_maker, vec
from tdlib import commander

def parse_vertex_names(s):
    """Convert a shorthand notation for vertex names to a list of pairs
    e.g. 'a 0 1 2 b 4' -> [('a',0),('a',1),('a',2),('b',4)]
    """
    pairs = []
    tokens = iter(s.split())
    sect = None
    for t in tokens:
        try:
            n = int(t)  # numeric = point number
            if sect:
                pairs.append((sect,n))
        except ValueError:  
            sect = t    # non-numeric = section name
    return pairs

def list_vertex_names(vnames):
    """Convert a list of vertex name pairs to shorthand notation"""
    tokens=[]
    if vnames:
        sect = None
        for vs,vn in vnames:
            if sect != vs:
                tokens.append(vs)
                sect = vs
            tokens.append(str(vn))
    return ' '.join(tokens)

default_primitive = {
    1:GL.GL_POINTS,
    2:GL.GL_LINES,
    3:GL.GL_TRIANGLES,
    4:GL.GL_QUADS
    }

primitive_name = {}
for name in [
    'POINTS','LINES','LINE_STRIP','LINE_LOOP',
    'TRIANGLES','TRIANGLE_STRIP','TRIANGLE_FAN',
    'QUADS','QUAD_STRIP','POLYGON'
    ]:
    primitive_name[getattr(GL,'GL_' + name,'?')] = name
    
class DictUndoer(object):
    """Object for handling undo/redo on dictionary-type objects"""
    def __init__(self,target,max_undo=100):
        self.target = target
        self.undolist = []
        self.redolist = []
        self.max_undo = max_undo
    def checkpoint(self):
        self.undolist.append(self.target.copy())
        if len(self.undolist) > self.max_undo:
            self.undolist.pop(0)
        self.redolist = [] # making a new checkpoint prevents redo
    def undo(self):
        if self.undolist:
            self.redolist.append(self.target.copy())
            undone_value = self.undolist.pop()
            self.target.clear()
            self.target.update(undone_value)
    def redo(self):
        if self.redolist:
            self.undolist.append(self.target.copy())
            redone_value = self.redolist.pop()
            self.target.clear()
            self.target.update(redone_value)
    def can_undo(self):
        return bool(self.undolist)
    def can_redo(self):
        return bool(self.redolist)

class AnyUndoer(DictUndoer):
    """Undo/redo manager for any attributes of an object."""
    __slots__ = ('target','undolist','redolist','max_undo')

    def getcopy(self,attributes):
        """Get a copy of one or more attributes of the target object
        as ([attribute_name],[pickled_attribute_data]) """
        pickles = [pickle.dumps(getattr(self.target,attr,None))
                 for attr in attributes]
        return (attributes,pickles)

    def restorecopy(self,attributes,pickles):
        """Put the pickled data back, replacing attributes of target object"""
        for i,attr in enumerate(attributes):
            setattr(self.target, attr, pickle.loads(pickles[i]))

    def checkpoint(self,attributes):
        """Save pickled copies of attributes (for a deep copy)"""
        self.undolist.append(self.getcopy(attributes))
        if len(self.undolist) > self.max_undo:
            self.undolist.pop(0)    # earliest saved state drops off
        self.redolist = []

    def undo(self):
        if self.undolist:
            attributes,pickles = self.undolist.pop()
            self.redolist.append(self.getcopy(attributes))
            self.restorecopy(attributes,pickles)
        else:
            attributes = ()
        return attributes   # changed attribute names

    def redo(self):
        if self.redolist:
            attributes,pickles = self.redolist.pop()
            self.undolist.append(self.getcopy(attributes))
            self.restorecopy(attributes,pickles)
        else:
            attributes = ()
        return attributes   # changed attribute names


class ViewerUpdater(object):
    """An object that records changes, and keeps a registry of callbacks
    to be applied when certain things have changed."""
    def __init__(self):
        self.registry = []
        self.changes = set()
    def register(self,callback, *care_about):
        """Specify what changes would require a function to be called"""
        entry = set(care_about), callback
        if entry not in self.registry:
            self.registry.append(entry)
    def unregister(self, callback, *care_about):
        """Remove the requirement to call a particular callback function"""
        entry = set(care_about), callback
        try:
            self.registry.remove(entry)
        except ValueError:
            pass
    def change(self,*args):
        """Record that one or more things have changed"""
        for item in args:
            self.changes.add(item)
    def update(self):
        """Call all the callbacks that need called for the changes so far"""
        if self.changes:
            for care_about, callback in self.registry:
                if self.changes & care_about:
                    callback()
            self.changes.clear()

default_treg = {
    'S0':[(0,0),(1,0),(1,1),(0,1)],
    '':[(0,0)],
    'A':[(0,0),(1,0),(0.5,1)]
    }

class SkeledModel(object):
    """An object holding the model information for the Skeled 3D model editor

    All changes to the model should be made by calling the methods here,
    and any changes that will need to be saved, or reflected in the user
    interface, will be noted correctly.

    2004-12-19 simplifying the API so all actions on faces are done on the
        selected set of faces, rather than having to get the selected faces
        and then pass them back in to specify which ones to change.
    """
    def __init__(self,updater):
        self.skel = tnv_maker.VertexSkeleton()
        self.selection = {}
        self.surfaces = {}
        self.treg = default_treg.copy()
        self.updater = updater
        self.undoer = AnyUndoer(self)
        self.unsaved = set()    # set of things needing to be saved
        self.clip_skel = {}
        self.clip_faces = {}
        self.scratch_coords = []
        self.chosen_pieces = set()
        self.chosen_faces = set()
        self.show_labels = True

    # Methods to record what needs to be saved, what has changed
    def need_to_save(self,*args):
        """ Add some things to the set of what needs to be saved """
        self.unsaved |= set(args)
        self.updater.change('save')
    def no_need_to_save(self,*args):
        """ Remove some things from the set of what needs to be saved """
        self.unsaved -= set(args)
        self.updater.change('save')

    # Methods for TNV data 
    def tnv_extract(self,pieces=None):
        """ Extract TNV data for a list of the pieces"""
        if not pieces:
            pieces = self.surfaces.keys()
        maker = tnv_maker.TNVmaker(self.skel,self.treg)
        for piece in pieces:
            facedict = self.surfaces.get(piece,{})
            faces = [(vs,r,t) for vs,(r,t) in self.face_items(piece)]
            if faces:
                maker.add_piece(piece, faces,
                                facedict.get('maptype','direct'),
                                facedict.get('primitive',None))
        maker.calculate_rounded()
        return maker.extract()
    def tnv_save(self,filename):
        """ Extract and save all TNV data"""
        tnv = self.tnv_extract()
        pickle.dump(tnv,file(filename,"wb",2))

    # Methods for texture regions
    def treg_load(self,filename,merge=False):
        """ Load texture region file"""
        try:
            treg = pickle.load(file(filename))
            assert type(treg) is dict
        except AssertionError:
            raise TypeError('Not a pickled dictionary')
        if not merge:
            self.treg.clear()
            self.no_need_to_save('treg')
        else:
            self.need_to_save('treg')
        self.treg.update(treg)
        self.updater.change('treg')
    def treg_save(self,filename):
        """ Save texture regions to a file"""
        pickle.dump(self.treg,file(filename,"wb"),2)
        self.no_need_to_save('treg')
        self.updater.change('treg')
    def treg_set(self,name,stcoords):
        """ Define a texture region by a list of (s,t) coordinate pairs"""
        self.treg[name] = stcoords
        self.need_to_save('treg')
        self.updater.change('treg')
    def treg_delete(self,name):
        """ Delete a texture region """
        try:
            del self.treg[name]
            self.need_to_save('treg')
            self.updater.change('treg')
        except KeyError:
            pass
    def treg_keys(self):
        """ Return a list of texture regions """
        return self.treg.keys()
    def treg_get(self,name):
        """ Get the (s,t) coordinate pairs for a named texture region """
        coords = self.treg.get(name,[])
        return coords[:]

    # Methods for scratch coords used to preview texture regions
    def scratch_set(self,stcoords):
        """ Set list of scratch coords """
        self.scratch_coords = stcoords[:]
        self.updater.change('scratch')
    def scratch_get(self):
        """ Get list of scratch coords """
        return self.scratch_coords[:]

    # Methods for faces
    def face_load(self,filename,merge=False):
        """Load faces from a file"""
        try:
            surf = pickle.load(file(filename))
            assert type(surf) is dict
        except AssertionError:
            raise TypeError('Not a pickled dictionary')
        if not merge:   # just replace all pieces
            self.surfaces.clear()
            self.no_need_to_save('face')
        else:
            self.need_to_save('face')
        self.surfaces.update(surf)
        self.updater.change('face')
    def face_save(self,filename):
        """Save faces to a file"""
        pickle.dump(self.surfaces,file(filename,'wb'),2)
        self.no_need_to_save('face')
        self.updater.change('face')

    def face_add_piece(self, piece, maptype='direct'):
        """Create a new piece, make it the currently selected one"""
        if piece in self.surfaces:
            return
        self.surfaces[piece] = {'maptype':maptype}
        self.chosen_pieces = set([piece])
        self.updater.change('face','choosepiece')
        self.need_to_save('face')

    def face_set_mapping(self, maptype='direct'):
        """Set texture mapping type of selected pieces"""
        for piece in self.chosen_pieces:
            if piece in self.surfaces:
                self.surfaces[piece]['maptype'] = maptype
            else:
                self.surfaces[piece] = {'maptype':maptype}
        self.updater.change('face','choosepiece')
        self.need_to_save('face')
    def face_get_mapping(self,piece,default):
        if piece in self.surfaces:
            return self.surfaces[piece].get('maptype',default)
        else:
            return default

    def face_set_primitive(self,piece,primitive=None):
        """Override the default guess of OpenGL primitive used to render the piece"""
        if type(primitive) == str:
            primitive = getattr(GL,'GL_' + primitive,None)
        for piece in self.chosen_pieces:
            if piece in self.surfaces:
                self.surfaces[piece]['primitive'] = primitive
            else:
                self.surfaces[piece] = {'primitive':primitive}
        self.updater.change('face','choosepiece')
        self.need_to_save('face')

    def face_get_primitive(self,piece):
        if piece in self.surfaces:
            primitive = self.surfaces[piece].get('primitive')
            if primitive is None:
                size = self.face_polygon_size(piece)
                return default_primitive.get(size, GL.GL_POLYGON)

    def face_get_chosen_piece(self):
        """Return name of single selected Piece, or None if 0 or more than 1"""
        if len(self.chosen_pieces) == 1:
            return list(self.chosen_pieces)[0]
        else:
            return None
    def face_get_chosen_pieces(self):
        """Return names of all selected pieces"""
        return list(self.chosen_pieces)

    def face_get_chosen_faces(self):
        """Return keys of all selected faces"""
        return list(self.chosen_faces)

    def face_choose_pieces(self,pieces):
        """Choose pieces by their names"""
        self.chosen_pieces = set(pieces) & set(self.surfaces.keys())
        if len(self.chosen_pieces) == 1:
            piece = list(self.chosen_pieces)[0]
            faces = self.surfaces.get(piece,{})
            self.chosen_faces &= set(faces.keys())
        else:
            self.chosen_faces.clear()
        self.updater.change('chooseface')
        self.updater.change('choosepiece')

    def face_choose_faces(self,faces,mode='replace'):
        """Choose faces by their keys"""
        piece = self.face_get_chosen_piece()    # one piece chosen
        if piece:
            piecefaces = self.surfaces.get(piece,{}) # can only choose faces in it
            chosen = set(faces) & set(piecefaces.keys())
            if mode in ('include','+'):
                self.chosen_faces |= chosen
            elif mode in ('toggle','^'):
                self.chosen_faces ^= chosen
            elif mode in ('refine','&'):
                self.chosen_faces &= chosen
            elif mode in ('exclude','-'):
                self.chosen_faces -= chosen
            else:
                self.chosen_faces = chosen
        else:
            self.chosen_faces.clear()
        self.updater.change('chooseface')

    def face_pieces(self):
        """List of all pieces in the model"""
        return self.surfaces.keys()

    def face_polygon_size(self,piece):
        """Get size of polygons of a piece, if any entered yet"""
        faces = self.face_items(piece)
        if faces:
            vertices, opts = faces[0]
            return len(vertices)
        else:
            return None

    def face_get(self,piece,key,default=None):
        """Get the roundedness and texture region of a face, or the value
            of a special entry in the faces dict"""
        faces = self.surfaces.get(piece,{})
        if type(key) is tuple:  # a face, key is vertex names
            default = (False,'')    # default is not round, no texture region
        return faces.get(key,default)

    def face_set(self,piece,key,value):
        """Set the roundedness and texture region of a face, or the value of
            a special entry in the faces dict"""
        faces = self.surfaces[piece]
        if type(key) is tuple:
            arity = None
            for k in faces.keys():
                if type(k) is tuple:
                    arity = len(k)
                    break
            if arity and len(key) != arity:
                raise ValueError('Wrong number of vertices')
            for v in key:
                if v not in self.skel:
                    raise KeyError('Vertex %s not in skeleton' % str(v))
        faces[key] = value
        self.updater.change('face')
        self.need_to_save('face')
        
    def face_gen_faces(self,pieces=None):
        """Generate a sequence of dicts containing face information"""
        if pieces is None:
            pieces = self.face_get_chosen_pieces()
            if not pieces:
                pieces = self.surfaces.keys()
        for piece in pieces:
            faces = self.surfaces[piece]
            primitive = self.face_get_primitive(piece)
            for key in faces:
                if type(key) is tuple:  # a face
                    d = {'piece':piece,
                        'names': list_vertex_names(key),
                        'coords': [self.skel.get(v) for v in key],
                        'primitive':primitive
                        }
                    yield d

    def face_set_faces(self,piece,keys,value):
        """Set multiple faces to have the same options"""
        for key in keys:
            self.face_set(piece,key,value)

    def face_delete_piece(self,piece):
        """Delete one piece of the model"""
        try:
            del self.surfaces[piece]
            self.updater.change('face')
            self.need_to_save('face')
            if piece in self.chosen_pieces:
                self.chosen_pieces.remove(piece)
                self.updater.change('choosepiece')
        except KeyError:
            pass

    def face_items(self,piece):
        """List of all faces of one piece of the model"""
        faces = self.surfaces.get(piece,{})
        return [(key,value)
                for key,value in faces.items()
                if type(key) is tuple]

    def face_delete_face(self,piece,face):
        """Delete one face of the piece"""
        try:
            del self.surfaces[piece][face]
            self.updater.change('face')
            self.need_to_save('face')
        except KeyError:
            pass
        if face in self.chosen_faces:
            self.chosen_faces.remove(face)
            self.updater.change('chooseface')

    def face_delete_faces(self,piece,faces):
        """Delete multiple faces of the piece"""
        for face in faces:
            self.face_delete_face(piece,face)

    def face_auto_mesh(self,piece,bucket_coord=2,bucket_tolerance=0.1,
                       rounded=False,texregion=''):
        """make a triangle mesh over the selected vertices, after
        gathering them into contours (groups with similar value for one
        of their coordinates).
        """
        faces = self.surfaces[piece]
        tmp = [0,1,2]
        tmp.remove(bucket_coord)
        coord1, coord2 = tmp
        buckets = []
        vsorted = []
        # pair selected vertex names with distance along chosen coord
        for k in self.selection:
            try:
                vertex = self.skel[k]
                vsorted.append((vertex[bucket_coord],k))
            except KeyError:
                continue
        vsorted.sort()  # in order of chosen coord
        if not vsorted:
            return
        # split sorted vertices into buckets wherever distance from one to the next
        # exceeds the tolerance
        prev = vsorted[0][0]    # lowest value of sorting coord
        bucket = []
        buckets.append(bucket)
        for curr, k in vsorted:
            if curr - prev > bucket_tolerance:
                bucket = []
                buckets.append(bucket)
            bucket.append(k)
            prev = curr
        # sort the vertices of each bucket by angle around some point
        contours = [tnv_maker.sort_by_angle(self.skel,bucket,coord1,coord2)
                    for bucket in buckets if bucket]
        # now contours are sorted rings of vertices
        contourA = contours[0]
        for contourB in contours[1:]:
            for v1,v2,v3 in tnv_maker.triangulator(self.skel,contourA,contourB,True):
                faces[v1,v2,v3] = (rounded,texregion)
            contourA = contourB
        self.updater.change('face')
        self.need_to_save('face')

    def face_subdivide(self,piece,face,convexity=0.0):
        """Subdivide a triangular face into 3 smaller ones, adding a new vertex"""
        if len(face) != 3:
            return
        choose = (face in self.chosen_faces)
        cx = cy = cz = 0.0
        coords=[]
        for key in face:
            vx, vy, vz = self.skel.get(key,(0.0,0.0,0.0))
            cx += vx
            cy += vy
            cz += vz
            coords.append((vx,vy,vz))
        cx /= 3
        cy /= 3
        cz /= 3
        nx,ny,nz = vec.plane_normal(*coords)
        section = face[0][0]
        new = self.skel_add(section,
                            (cx + convexity * nx,
                             cy + convexity * ny,
                             cz + convexity * nz))
        v0,v1,v2 = face
        opts = self.surfaces[piece].get(face,(False,''))
        newfaces = [(v0,v1,new),(v1,v2,new),(v2,v0,new)]
        if choose:
            self.face_choose_faces(newfaces,mode='include')
        for newface in newfaces:
            self.face_set(piece,newface,opts)
        self.face_delete_face(piece,face)        

    def face_rename_piece(self,piece,newname):
        """Re-name a piece of the model"""
        try:
            self.surfaces[newname] = self.surfaces[piece]
            self.face_delete_piece(piece)
        except KeyError:
            pass

    # Methods for skeleton vertices
    def skel_clear(self):
        """Remove all vertices"""
        self.skel.clear()
        self.select_none()
        self.updater.change('skel')
        self.no_need_to_save('skel')
    def skel_load(self,filename,merge=False):
        """Load vertices from a file"""
        self.skel.load(filename,merge)
        self.updater.change('skel')
        if not merge:
            self.select_none()
            self.updater.change('selection')
            self.no_need_to_save('skel')
        else:
            self.need_to_save('skel')
    def skel_save(self,filename):
        """Save vertices to a file"""
        self.skel.save(filename)
        self.no_need_to_save('skel')
    def skel_add(self,section,coords):
        """Add a vertex in the named section"""
        new = self.skel.add(section,coords)
        self.updater.change('skel')
        self.need_to_save('skel')
        return new
    def skel_items(self):
        return self.skel.items()
    def skel_get(self,key):
        return self.skel[key]

    # Methods for undo/redo
    def checkpoint(self,*attributes):
        self.undoer.checkpoint(attributes)
        self.updater.change('undo')
    def undo(self):
        attributes = self.undoer.undo()
        self.updater.change('undo',*attributes)
    def redo(self):
        attributes = self.undoer.redo()
        self.updater.change('undo',*attributes)
    def can_undo(self):
        return self.undoer.can_undo()
    def can_redo(self):
        return self.undoer.can_redo()

    # Methods for the selection
    def selection_nonempty(self):
        return bool(self.selection)

    def select_none(self):
        self.selection.clear()
        self.updater.change('selection')

    def select_all(self):
        self.selection.clear()
        for k in self.skel:
            self.selection[k] = 1.0
        self.updater.change('selection')

    def select_by_criteria(self,names=None,
                           X=None,Y=None,Z=None,R=1.0,
                           Xop='=',Yop='=',Zop='=',Rop='<',
                           mode='replace'):
        """Select skeleton points according to search criteria"""
        sel = {}
        if not names:
            keys = self.skel.keys()
        else:
            keys = parse_vertex_names(names)
            if not keys:
                keys = [ k for k in self.skel if k[0] in names]
        for key in keys:
            try:
                x,y,z = self.skel[key]
            except KeyError:
                continue
            if (X is None or
                (x < X and Xop == '<') or
                (x > X and Xop == '>')):
                x0 = x
            else:
                x0 = X
            if (Y is None or
                (y < Y and Yop == '<') or
                (y > Y and Yop == '>')):
                y0 = y
            else:
                y0 = Y
            if (Z is None or
                (z < Z and Zop == '<') or
                (z > Z and Zop == '>')):
                z0 = z
            else:
                z0 = Z
            d = sqrt((x0-x)**2 + (y0-y)**2 + (z0-z)**2)
            if Rop == '<':
                if d > R:
                    continue
                else:
                    try:
                        closeness = 1.0 - (d/R)
                    except ZeroDivisionError:
                        closeness = 0.0
            elif Rop == '>':
                closeness = float(d > R)
            if closeness:
                sel[key] = closeness
        self.select_these(sel,mode)

    def select_these(self,sel,mode='replace'):
        if not isinstance(sel,dict):
            sel = dict.fromkeys(sel,1.0)
        allvertices = set(self.skel.keys())
        if mode in ('replace', ''):
            self.selection.clear()
        if mode in ('replace','','include','+'):
            self.selection.update(sel)
        elif mode in ('exclude', '-'):
            for k in sel:
                try:
                    del self.selection[k]
                except KeyError:
                    pass
        elif mode in ('refine', '&'):
            for k,amount in self.selection.items():
                factor = sel.get(k,0.0)
                amount *= factor
                if not amount:
                    del self.selection[k]
                else:
                    self.selection[k] = amount
        # don't select vertices that don't exist in the skeleton
        for k in self.selection:
            if k not in allvertices:
                del self.selection[k]
        self.updater.change('selection')

    def select_keys(self):
        return self.selection.keys()
    def select_items(self):
        return self.selection.items()
    def select_get(self,key):
        return self.selection.get(key,0.0)

    # Cut, copy, paste
    def select_copy(self,withfaces=True):
        clip_skel = {}
        clip_faces = {}
        for k in self.selection:
            clip_skel[k] = self.skel[k]
        if withfaces:
            for piece in self.surfaces:
                faces = self.face_items(piece)
                clip_faces[piece] = {}
                for vertices,opts in faces:
                    out = False
                    for v in vertices:
                        if v not in self.selection:
                            out = True
                            break
                    if not out:
                        clip_faces[piece][vertices] = opts
        self.clip_skel = clip_skel
        self.clip_faces = clip_faces
        self.updater.change('clipboard')
    def select_delete(self):
        for k in self.selection:
            try:
                del self.skel[k]
            except KeyError:
                pass
        # NB - don't delete faces!  They take too much effort to replace
        if self.selection:
            self.updater.change('skel')
            self.need_to_save('skel')
        self.selection.clear()
        self.updater.change('selection')        
    def select_cut(self,withfaces=True):
        self.select_copy(withfaces)
        self.select_delete()
    def select_paste(self,section=None,withfaces=True):
        skel_lookup={}
        for (sect,num),coords in self.clip_skel.items():
            if section is None:
                newsect = sect
            else:
                newsect = section
            newkey = self.skel.add(newsect,coords)
            skel_lookup[sect,num] = newkey
        self.select_these(skel_lookup.values(),'replace')
        if withfaces:
            for piece,faces in self.clip_faces.items():
                if (piece not in self.surfaces) and faces:
                    self.surfaces[piece] = {}
                for vertices,opts in faces.items():
                    newvertices = tuple([ skel_lookup.get(v,v) for v in vertices])
                    self.surfaces[piece][newvertices] = opts
            if self.clip_faces:
                self.updater.change('face')
                self.need_to_save('face')
    def select_rename(self,section):
        """Rename selected vertices into a different section, change
        faces to use new vertex names"""
        skel_lookup = {}
        for k in self.selection:
            skel_lookup[k] = self.skel_add(section,self.skel[k])
            del self.skel[k]
        self.selection.clear()
        for piece,faces in self.surfaces.items():
            for vertices,opts in self.face_items(piece)[:]:
                vlist = [ skel_lookup.get(v,v) for v in vertices ]
                del faces[vertices]
                faces[tuple(vlist)] = opts
        for k in skel_lookup:
            self.selection[k] = 1.0
        self.updater.change('skel','face','selection')
        self.need_to_save('skel','face')
    def select_split_piece(self,piece):
        """Split a piece into 2 pieces, one with faces only containing
        selected vertices, and one with faces containing at least one
        unselected vertex"""
        if piece not in self.surfaces:
            return  # piece doesn't exist
        n = 0
        selpiece = "%s%d" % (piece,n)   # name for split-off piece
        while selpiece in self.surfaces:
            n += 1
            selpiece = "%s%d" % (piece,n)
        selfaces = {}   # faces only with selected vertices
        unselfaces = {} # other faces
        for vertices,opts in self.face_items(piece):
            allselected = min([(v in self.selection) for v in vertices])
            if allselected:
                selfaces[vertices] = opts
        if selfaces:
            self.surfaces[selpiece] = selfaces
            faces = self.surfaces[piece]
            for vertices in selfaces:
                del faces[vertices]  # remove from other piece
            self.updater.change('face')
            self.need_to_save('face')
    def can_paste(self):
        """Return 0 if nothing to paste, 1 if vertices, 2 if faces also"""
        return bool(self.clip_skel) + bool(self.clip_faces)

    # Transforms on the selection
    def select_translate(self,dx,dy,dz):
        for key, amount in self.selection.items():
            vx,vy,vz = self.skel[key]
            vx += dx * amount
            vy += dy * amount
            vz += dz * amount
            self.skel[key] = (vx,vy,vz)
        self.updater.change('skel')
        self.need_to_save('skel')

    def select_scale(self,dx,dy,dz):
        for key, amount in self.selection.items():
            vx,vy,vz = self.skel[key]
            vx *= dx * amount
            vy *= dy * amount
            vz *= dz * amount
            self.skel[key] = (vx,vy,vz)
        self.updater.change('skel')
        self.need_to_save('skel')

    def select_rotate(self,deg,axisfrom,axisto):
        """Use quaternions to rotate selected vertices around
        an axis specified by an arbitrary line segment"""
        x0,y0,z0 = axisfrom
        x1,y1,z1 = axisto
        i,j,k = vec.normalise(x1-x0,y1-y0,z1-z0)
        halfangle = radians(deg / 2.0)
        c = cos(halfangle)
        s = sin(halfangle)
        q = quaternion(c, i*s, j*s, k*s)
        for key in self.selection.keys():
            x,y,z = self.skel_get(key)
            vrot = quaternion(0,x-x0,y-y0,z-z0)  # translate origin to axisfrom
            new = q * vrot * q.conj()    # the rotation
            newcoords = new.i + x0, new.j + y0, new.k + z0 # translate back
            self.skel[key] = newcoords
        self.updater.change('skel')
        self.need_to_save('skel')

    # Settings only used for presentation, irrelevant to the model
    def toggle_labels(self):
        self.show_labels = not(self.show_labels)
        self.updater.change('labels')

class SkeledCommander(commander.CommandInterpreter):
    """Command-line control of a skeled model
    Commands are:
    / vertex names /    :   select vertices directly
        May be followed by -,+,& to exclude, add or intersect with selection
    //=                 :    list selected vertices
    ? expression ?      :  select vertices by python expression
        May be followed by -,+,& as for /.../ command
        variables set for expression are:
            name    : name part of vertex label
            num     : num part of vertex label
            x ,y, z : coords of vertex
            F       : number of faces that include the vertex
            f       : number of selected faces that include the vertex
            X, Y, Z : centre coordinates (initially 0,0,0)
            i, j, k : axis components (initially 0,0,1)
            r       : distance of vertex from centre
    centre x,y,z    : set centre to x,y,z
    centre          : show centre coords
    zoom n          : set zoom to n
    zoom            : show zoom
    axis i,j,k      : set axis to i,j,k
    axis X|Y|Z      : set axis to X, Y or Z axis
    . x,y,z name    : add a vertex at x,y,z label with name (name optional)

    ... others to be decided...

    """
    def __init__(self,model,*args,**kw):
        super(SkeledCommander,self).__init__(*args,**kw)
        self.model = model
        self.updater = model.updater    # temporary
        self.undoer = model.undoer      # temporary
        self.section_name = 'v'
        self.centre = (0,0,0)
        self.axis = (0,0,1)
        self.default_help_command()
        self.autoplug()

    def vertex_expr_dict(self):
        X,Y,Z = self.centre
        i,j,k = self.axis
        variables = dict(X=X, Y=Y, Z=Z, i=i, j=j, k=k,
            True=True,False=False)
        return variables

    def list_selection(self):
        sel = self.model.selection.keys()
        sel.sort() # sort by section makes list more compact
        return list_vertex_names(sel)

    def cmd_vertex(self, coords='', name=''):
        r"""
        <help>
        . x,y,z [name] : add a vertex
        </help>
        <pattern>
            (?P<coords>[^,]*,[^,]*,[^ ]*)
            \s*
            (?P<name>[A-Za-z]*)
        </pattern>

        Add a vertex to named section of skeleton, at specified coords
        """
        if not name:
            name = self.section_name
        variables = self.vertex_expr_dict()
        evalcoords = commander.safe_eval(coords,variables)
        self.section_name = name
        return self.model.skel_add(name,evalcoords)

    def cmd_select_names(self,names='',mode=''):
        r"""
        <help>
        / vertex names / : select vertices
        /*/ selects all, /.../+ add, /.../- remove, /.../& restrict
        </help>
        <pattern>
            (?P<names>[^/]*)/(?P<mode>\+|-|&|=|)
        </pattern>

        Select vertices by vertex names
        """
        if mode == '=':    # no change, just show selection
            pass
        else:
            if names == '*':
                sel = self.model.skel.keys()
            else:
                sel = parse_vertex_names(names)
            self.model.select_these(sel, mode=mode)
        # show results
        return '/%s/' % self.list_selection()

    def cmd_select_expr(self,expr='',mode=''):
        r"""
        <help> 
        ? expr ? : select vertices by logical expression
        </help>
        <pattern>
            (?P<expr>[^?]*)\?(?P<mode>\+|-|&|)
        </pattern>

        Select vertices by logical expression
        """
        # variables for the expression, not depending on vertex
        variables = self.vertex_expr_dict()        
        # do the selection
        these = {}
        for ((name,num),(x,y,z)) in self.model.skel.items():
            vertexvars = dict(x=x, y=y, z=z, name=name, num=num)
            vertexvars.update(variables)
            selected = commander.safe_eval(expr,vertexvars)
            if selected:
                these[name,num] = selected
        self.model.select_these(these, mode=mode)
        # show results
        return '/%s/' % self.list_selection()

def main():
    import sys
    updater = ViewerUpdater()
    com = SkeledCommander(SkeledModel(updater))
    com.output = sys.stdout
    while True:
        line = raw_input('Skeled> ')
        if line == 'quit':
            break
        try:
            output = com.execute(line)
        except commander.NotRecognised, n:
            print 'Error: %s' % n.args

if __name__ == '__main__':
    main()

