#! /usr/bin/env python
"""A Texture is a manager of an OpenGL tecture object, constructed from
    a pygame surface.

    Best way to get a texture is to call get_texture(filename), which will re-use
    the same Texture object if it is called again while the object is still in use.
    
This file copyright Peter Harris 2004, released under the terms of the
GNU GPL. See www.gnu.org for details.

"""

from OpenGL import GL
from OpenGL.GLU import gluBuild2DMipmaps
from pygame import image
from weakref import WeakValueDictionary

texture_pool = WeakValueDictionary()

def enable():
    """Convenience function to glEnable(GL_TEXTURE_2D)"""
    GL.glEnable(GL.GL_TEXTURE_2D)

def disable():
    """Convenience function to glDisable(GL_TEXTURE_2D)"""
    GL.glDisable(GL.GL_TEXTURE_2D)

def is_a_power_of_2(x):
    """Binary trick: x & -x yields lowest set bit in x.
    For powers of 2 it's the only bit"""
    return x == (x & -x)
    
class Texture:
    def __init__(self,filename):
        """Create texture from image file"""
        self.tname = GL.glGenTextures(1)
        img = image.load(filename)
        width,height = img.get_size()
        if not (is_a_power_of_2(width) and is_a_power_of_2(height)):
            print "Warning: size of texture image %s is (%d,%d)", (filename,width,height)
        img_data = image.tostring(img,"RGBA",True)
        GL.glBindTexture(GL.GL_TEXTURE_2D,self.tname)
        gluBuild2DMipmaps(GL.GL_TEXTURE_2D, GL.GL_RGBA8,
                        width, height, GL.GL_RGBA, GL.GL_UNSIGNED_BYTE,
                        img_data)
        alphabytes = img_data[3::4]
        if alphabytes != '\xff' * len(alphabytes):  # some transparency
            self.transparent = True
        else:
            self.transparent = False

    def bind(self):
        GL.glBindTexture(GL.GL_TEXTURE_2D,self.tname)

    def __del__(self):
        if GL:
            GL.glDeleteTextures([self.tname])
        
def get_texture(filename):
    """Get a texture object or re-use one already loaded"""
    if filename not in texture_pool:
        tex = Texture(filename)
        texture_pool[filename] = tex
    return texture_pool[filename]
