#! /usr/bin/python
""" A PerspectiveGroup sets up a perspective transformation before drawing its
    contents.

    A ViewGroup is a  perspective group that also sets up lighting and camera
    position.
    
    Lighting used to be the responsibility of a LightingRig object, but because
    different objects might need to acquire lights for special effects, lighting
    was moved to its own module.
    Likewise the camera has been factored out to a camera module.

    Copyright 2005 Peter Harris
    Released under the terms of the GNU General Public License
    (See www.gnu.org for details)

"""
from __future__ import division


from OpenGL.GL import *
from OpenGL.GLU import gluPerspective

import pygame

from tdlib import part, lighting, camera, picking

class ProjectionGroup(part.Group):
    """An abstract class - need to subclass with an __init__ that
       calls build_dlist after setting up all details needed
       """
    def __init__(self,*args,**kwd):
        super(ProjectionGroup,self).__init__(*args,**kwd)
    def build_dlist(self):
        """Make a display list to perform the projection.
        This may be a performance improvement over doing it directly each time.
        """
        self.dl_projection = glGenLists(1)
        glNewList(self.dl_projection,GL_COMPILE)
        self.project()
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glEnable(GL_DEPTH_TEST)
        glEndList()
    def __del__(self):
        """Clean up display list"""
        glDeleteLists(self.dl_projection,1)
    def setup_geom(self):
        picking.project()   # set up identity or picking projection matrix
        glCallList(self.dl_projection)
    def setdown_geom(self):
        pass
    def project(self):
        pass

class OrthoGroup(ProjectionGroup):
    _default_geom = dict(left=-1.0,right=1.0,top=1.0,bottom=-1.0,near=-1.0,far=100.0)
    def __init__(self,*args,**kwd):
        super(OrthoGroup,self).__init__(*args,**kwd)
        self.build_dlist()
    def project(self):
        near = self.getgeom('near',-1.0)
        far = self.getgeom('far',100.0)
        left = self.getgeom('left',-1.0)
        right = self.getgeom('right',1.0)
        top = self.getgeom('top',1.0)
        bot = self.getgeom('bottom',-1.0)
        glOrtho(left, right, bot, top, near, far)
    def __del__(self):
        super(OrthoGroup,self).__del__()

class PerspectiveGroup(ProjectionGroup):
    """A group that sets up perspective when it is rendered"""
    _default_geom = {
        'perspective_angle':15.0,
        'near':1.0,
        'far':1000.0 }
    def __init__(self, *args, **kwd):
        super(PerspectiveGroup,self).__init__(*args,**kwd)
        try:
            size = self._geom['size']
        except KeyError:
            size = pygame.display.get_surface().get_size()
            self._geom['size'] = size
        w,h = size
        if h == 0: h = 1
        self.aspect = w/h
        self.near = self.getgeom('near',1.0)
        self.far = self.getgeom('far',1000.0)
        self.perspective_angle = self.getgeom('perspective_angle',15.0)
        self.build_dlist()
    def project(self):
        gluPerspective(
            self.perspective_angle,
            self.aspect,
            self.near,
            self.far)
    def __del__(self):
        super(PerspectiveGroup,self).__del__()

class ViewGroup(PerspectiveGroup):
    _default_geom = {
        'perspective_angle':15.0,
        'near':1.0,
        'far':1000.0,
        }
    _default_style = {
        'light-pos':(0,0,0,1),
        'light-colour':(1,1,1,0),
        'light-direction':(0,0,-1),
        'light-cutoff':180.0,
        }
    _style_attributes = ('light-pos','light-colour')
    
    def __init__(self, *args, **kwd):
        super(ViewGroup,self).__init__(*args,**kwd)
        self.light = lighting.claim_light()
        lighting.two_side(True)
        lighting.light_attenuation(self.light,0.1)
        lighting.light_switch(self.light,True)
        self.prepare()
        
    def light_position(self,pos,steps=0):
        lighting.light_position(self.light,pos,steps)
    
    def light_colour(self,colour,steps=0):
        lighting.light_colour(self.light,colour,steps)

    def spot_direction(self,direction,steps=0):
        lighting.light_spot_direction(self,direction,steps)
        
    def spot_cutoff(self,cutoff,steps=0):
        lighting.light_spot_cutoff(self,cutoff,steps)

    def setup_geom(self):
        """call camera and lighting setup after perspective transformation"""
        super(ViewGroup,self).setup_geom()
        camera.setup()

    def setup_style(self):
        lighting.setup()
        
    def prepare(self):
        light_pos = self.getstyle('light-pos',(0,0,0,1))
        light_colour = self.getstyle('light-colour',(1,1,1,0))
        lighting.light_position(self.light,light_pos)
        lighting.light_colour(self.light,light_colour)
    
    def step(self):
        """Let camera and lighting updates happen"""
        camera.step()
        lighting.update_lights()
        self.step_contents()
        
    def __del__(self):
        """Don't need the light any more"""
        super(ViewGroup,self).__del__()  # need to release display list
        lighting.release_light(self.light)
