"""
Tornado Test Viewer
"""
import pygame
from OpenGL import GL
from math import sin,cos,radians,fmod
from functools import partial

from tdlib import (part, viewgroup, tnvpart, mainloop,
                   camera, lighting, stylesheet, backdrop)
from tdlib import usualGL

import data
import tornado, sky, objpart, racetrack

BackdropStyle = {}
twisterstyle = {'texture':'twister.png', 'alpha':0.9}
skystyle = {'texture':'stormy.jpg'}
objstyle = {'obj-filename':'box.obj'}
lightStyle = {
    'light-pos':(40,-40,20,0),
    'light-colour':(0.5,0.5,0.3,1)
    }

class TornadoViewer(part.Group):
    def __init__(self,*args,**kwd):
        """Create a viewgroup with a Tornado inside it, let user
        look at it from various angles."""
        super(TornadoViewer,self).__init__(*args,**kwd)
        self.r = 20.0
        self.theta = 270.0
        self.height = 8.0
        camera.look_at((0,0,0))
        camera.look_from((0,-50,self.height))
        twister = tornado.Tornado('twister')
        skyball = sky.SkyBall('sky')
        obj = objpart.ObjPart('obj')
        track = racetrack.RaceTrack("tracks/test.track")
        scape = racetrack.Landscape('track',track)
        obj.pos = (6,0,track.z_at(6,0))
        vgroup = viewgroup.ViewGroup('view',[skyball, twister, obj, scape])
        # light-colour mistakenly sets ambient = diffuse. Need less ambient
        #lighting.adjust_light(vgroup.light,GL.GL_AMBIENT,(0.3,0.3,0.3,1.0))
        self.append(vgroup)
        self.keycommands = {
            pygame.K_LEFT:partial(self.track_left_or_right,30),
            pygame.K_RIGHT:partial(self.track_left_or_right,-30),
            pygame.K_UP:partial(self.track_up_or_down,5),
            pygame.K_DOWN:partial(self.track_up_or_down,-5),
            pygame.K_EQUALS:partial(self.zoom_in_or_out,0.75),
            pygame.K_MINUS:partial(self.zoom_in_or_out,1.5),
            }
    def move_viewpoint(self):
        camera.look_from((self.r * cos(radians(self.theta)),
                          self.r * sin(radians(self.theta)), self.height), 20)
    def track_left_or_right(self,amount):
        self.theta = fmod(self.theta + amount,360)
        self.move_viewpoint()
    def track_up_or_down(self,amount):
        self.height += amount
        self.move_viewpoint()
    def zoom_in_or_out(self,amount):
        self.r = min(max(0.2,self.r * amount),480.0)
        self.move_viewpoint()
    def event(self,event):
        """Change angle of view when left or right keys pressed."""
        if event.type == pygame.QUIT:
            self._expired = True
        if event.type == pygame.KEYDOWN:
            key = event.key
            if key == pygame.K_ESCAPE:
                self.quit = True
            elif key in self.keycommands:
                self.keycommands[key]()
        self.update_contents(event)
        return True    

def main():
    import sys
    import os
    args = iter(sys.argv[1:])
    tnv_filename=''
    tex_filename=None
    obj_filename=None
    pieces = []
    size = (800,600)
    bgcolour = (0.2,0.3,0.4,1.0)
    colour = (1,1,1,1)
    for arg in args:
        if arg == '--bg':
            bgcolour = tuple([float(s) for s in args.next().split(',')])
        else:
            name,ext = os.path.splitext(arg)
            if ext in ('.png','.jpeg','.jpg'):
                tex_filename = arg
            elif ext == '.obj':
                obj_filename = arg
    mainloop.open_window(size)
    usualGL.usual_setup()
    lighting.local_viewer(True)
    if tex_filename:
        twisterstyle['texture'] = tex_filename
    if obj_filename:
        objstyle['obj-filename'] = obj_filename
    BackdropStyle['background-colour'] = bgcolour
    
    stylesheet.load(
        {'Tornado':twisterstyle,
         'ObjPart':objstyle,
         'SkyBall':skystyle,
         'ViewGroup':lightStyle,
         'Backdrop':BackdropStyle})
    viewer = TornadoViewer('',[])
    mainloop.until_expired(viewer)

if __name__=='__main__':
    main()
