"""
    Tornado

    A tornado object

    Pyweek 5 Scavengers team entry  Sep 2007

    GPL v3 or later
"""
from data import filepath
from tdlib import part, texture, vec,animator
from OpenGL.GL import *
from math import sin, cos, radians, pi, fmod, atan2
from random import random

class Tornado(part.Part):
    """ A tornado
        pos is position of centre of top section of tornado.

        Tornado is drawn as a series of skewed cone sections, joining
        rings of points with radius increasing towards the top.

        Each ring is offset from the centre line by a distance and angle.

    """
    _default_geom = {
        'pos':(0.0,0.0,0.0),
        'slices':20,
        'heights':[0.0,1.0,2.0,8.0,12.0,15.0,30.0,100.0],
        'radii':[1.0,1.2,1.5,2.5,4.0,5.0,8.0,20.0],
        'distances':[1.0, 0.5, 1.0, 1.0, 1.8, 2.0, 5.0, 0.0],
        'angles':[0.0, 0.0, 180.0, 90.0, 100.0, 70.0, -30.0, 0.0],
        }
    _default_style = {'texture':'twister.png','colour':(1,1,1,0.9)}
    _style_attributes = ('texture','colour')
    _has_transparent = True
    def __init__(self, name="twister", track=None, geom=None, style=None, **kw):
        super(Tornado,self).__init__(name,geom=geom,style=style,**kw)
        self.dlist = None
        self.tex = None
        self.twist = 0.0
        self.twist_rate = -0.03
        self.track = track
        self.mutator = animator.Mutator(self._geom)
        if track:
            self.random_move()
    def prepare(self):
        if self.dlist is None:
            self.dlist = glGenLists(1)
        getgeom = self.getgeom
        slices = getgeom('slices',20)
        rings = []
        norms = [(cos(radians(a)), sin(radians(a)))
                 for a in range(0,360,360//slices)]
        geom = zip(getgeom('heights'),
                   getgeom('radii'),
                   getgeom('distances'),
                   map(radians, getgeom('angles')))
        for height, radius, dist, angle in geom:
            cx = dist * cos(angle)
            cy = dist * sin(angle)
            rings.append([
                (radius * nx + cx,
                 radius * ny + cy, height)
                for nx,ny in norms])
        topz = geom[-1][0]

        glNewList(self.dlist, GL_COMPILE)
        for i in range(len(rings) - 1):
            x,y,z0 = rings[-1][0]
            glBegin(GL_TRIANGLE_STRIP)
            s = 0.0
            for j in range(-1, slices):
                nx,ny = norms[j]
                glNormal3f(nx,ny,0.0)
                x,y,z = rings[i][j]
                x1,y1,z1 = rings[i+1][j]
                t = z / z0
                glTexCoord2f(s, t)
                glVertex3f(x,y,z)
                t1 = z1 / z0
                glTexCoord2f(s, t1)
                glVertex3f(x1,y1,z1)
                s += 1.0 / slices
            glEnd()
        glEndList()

        texfile = self.getstyle("texture")
        if texfile:
            self.tex = texture.get_texture(filepath(texfile))
        self.colour = self.getstyle("colour",(1,1,1,0.9))

    def setup_style(self):
        if self.tex:
            glEnable(GL_TEXTURE_2D)
            glTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
            glMatrixMode(GL_TEXTURE)
            glPushMatrix()
            glLoadIdentity()
            glTranslatef(self.twist,0.0,0.0)
            glMatrixMode(GL_MODELVIEW)
            self.tex.bind()
        else:
            glDisable(GL_TEXTURE_2D)
        glColor4fv(self.colour)
        glEnable(GL_BLEND)

    def setdown_style(self):
        if self.tex:
            glMatrixMode(GL_TEXTURE)
            glPopMatrix()
            glMatrixMode(GL_MODELVIEW)

    def render(self,mode):
        glEnable(GL_CULL_FACE)
        glEnable(GL_BLEND)
        glCullFace(GL_FRONT)
        glCallList(self.dlist)
        glCullFace(GL_BACK)
        glCallList(self.dlist)

    def step(self):
        self.twist = fmod(self.twist + self.twist_rate, 1.0)
        if self.track:
            self.mutator.step()
            if self.mutator.finished():
                self.random_move()
        changed = False
        getgeom = self.getgeom
        stacks = len(getgeom('heights'))
        angles = getgeom("angles")
        distances = getgeom("distances")
        for i in range(stacks):
            if random() < 0.1:
                changed = True
                angles[i] += random() * 10
        if changed:
            self.prepare()

    def random_move(self):
        track = self.track
        x = (random() - 0.5) * track.width
        y = (random() - 0.5) * track.length
        x0,y0,z0 = self.pos
        d = abs(x-x0) + abs(y-y0)
        self.mutator.change("pos",(x,y,0.0),int(d*30))
        
    def base_centre(self):
        dist = self.getgeom('distances')[0]
        a = radians(self.getgeom('angles')[0])
        px,py,pz = self.pos
        return px + dist*cos(a), py + dist*sin(a), pz

    def wind_vec(self,x,y,z):
        base = vec.Vec(self.base_centre())
        if z < 16.0: # if near enough, ignore vertical distance
            z = base.z
        target = vec.Vec(x,y,z)
        diff = target - base
        dsquared = max(diff.dot(diff),1.0)
        speed = 100.0 / dsquared
        # anti-clockwise
        wind = diff.orthoXY().normalise()
        # inward suction increases linearly toward the centre
        k = 1.0 / dsquared**0.5
        wind = wind*(1.0-k) + (diff.normalise() * -k)
        if dsquared <= 4.0:
            wind.z = 0.4 # fling upwards
        return wind * speed
                                
